from aerosandbox.dynamics.rigid_body.common_rigid_body import _DynamicsRigidBodyBaseClass
import aerosandbox.numpy as np
from aerosandbox.weights.mass_properties import MassProperties
from typing import Union


class DynamicsRigidBody3DBodyEuler(_DynamicsRigidBodyBaseClass):
    """
    Dynamics instance:
    * simulating a rigid body
    * in 3D
    * with velocity parameterized in body axes
    * and angle parameterized in Euler angles
    
    State variables:
        x_e: x-position, in Earth axes. [meters]
        y_e: y-position, in Earth axes. [meters]
        z_e: z-position, in Earth axes. [meters]
        u_b: x-velocity, in body axes. [m/s]
        v_b: y-velocity, in body axes. [m/s]
        w_b: z-velocity, in body axes. [m/s]
        phi: roll angle. Uses yaw-pitch-roll Euler angle convention. [rad]
        theta: pitch angle. Uses yaw-pitch-roll Euler angle convention. [rad]
        psi: yaw angle. Uses yaw-pitch-roll Euler angle convention. [rad]
        p: x-angular-velocity, in body axes. [rad/sec]
        q: y-angular-velocity, in body axes. [rad/sec]
        r: z-angular-velocity, in body axes. [rad/sec]

    Control variables:
        Fx_b: Force along the body-x axis. [N]
        Fy_b: Force along the body-y axis. [N]
        Fz_b: Force along the body-z axis. [N]
        Mx_b: Moment about the body-x axis. [Nm]
        My_b: Moment about the body-y axis. [Nm]
        Mz_b: Moment about the body-z axis. [Nm]
        hx_b: Angular momentum (e.g., propellers) about the body-x axis. [kg*m^2/sec]
        hy_b: Angular momentum (e.g., propellers) about the body-y axis. [kg*m^2/sec]
        hz_b: Angular momentum (e.g., propellers) about the body-z axis. [kg*m^2/sec]

    """

    def __init__(self,
                 mass_props: MassProperties = None,
                 x_e: Union[float, np.ndarray] = 0,
                 y_e: Union[float, np.ndarray] = 0,
                 z_e: Union[float, np.ndarray] = 0,
                 u_b: Union[float, np.ndarray] = 0,
                 v_b: Union[float, np.ndarray] = 0,
                 w_b: Union[float, np.ndarray] = 0,
                 phi: Union[float, np.ndarray] = 0,
                 theta: Union[float, np.ndarray] = 0,
                 psi: Union[float, np.ndarray] = 0,
                 p: Union[float, np.ndarray] = 0,
                 q: Union[float, np.ndarray] = 0,
                 r: Union[float, np.ndarray] = 0,
                 ):
        # Initialize state variables
        self.mass_props = MassProperties() if mass_props is None else mass_props
        self.x_e = x_e
        self.y_e = y_e
        self.z_e = z_e
        self.u_b = u_b
        self.v_b = v_b
        self.w_b = w_b
        self.phi = phi
        self.theta = theta
        self.psi = psi
        self.p = p
        self.q = q
        self.r = r

        # Initialize control variables
        self.Fx_b = 0
        self.Fy_b = 0
        self.Fz_b = 0
        self.Mx_b = 0
        self.My_b = 0
        self.Mz_b = 0
        self.hx_b = 0
        self.hy_b = 0
        self.hz_b = 0

    @property
    def state(self):
        return {
            "x_e"  : self.x_e,
            "y_e"  : self.y_e,
            "z_e"  : self.z_e,
            "u_b"  : self.u_b,
            "v_b"  : self.v_b,
            "w_b"  : self.w_b,
            "phi"  : self.phi,
            "theta": self.theta,
            "psi"  : self.psi,
            "p"    : self.p,
            "q"    : self.q,
            "r"    : self.r,
        }

    @property
    def control_variables(self):
        return {
            "Fx_b": self.Fx_b,
            "Fy_b": self.Fy_b,
            "Fz_b": self.Fz_b,
            "Mx_b": self.Mx_b,
            "My_b": self.My_b,
            "Mz_b": self.Mz_b,
            "hx_b": self.hx_b,
            "hy_b": self.hy_b,
            "hz_b": self.hz_b,
        }

    def state_derivatives(self):
        """
        Computes the state derivatives (i.e. equations of motion) for a body in 3D space.

        Based on Section 9.8.2 of Flight Vehicle Aerodynamics by Mark Drela.

        Returns:
            Time derivatives of each of the 12 state variables, given in a dictionary:
                {
                    "xe"   : d_xe,
                    "ye"   : d_ye,
                    "ze"   : d_ze,
                    "u"    : d_u,
                    "v"    : d_v,
                    "w"    : d_w,
                    "phi"  : d_phi,
                    "theta": d_theta,
                    "psi"  : d_psi,
                    "p"    : d_p,
                    "q"    : d_q,
                    "r"    : d_r,
                }
        """
        ### Shorthand everything so we're not constantly "self."-ing:
        u = self.u_b
        v = self.v_b
        w = self.w_b
        phi = self.phi
        theta = self.theta
        psi = self.psi
        p = self.p
        q = self.q
        r = self.r
        X = self.Fx_b
        Y = self.Fy_b
        Z = self.Fz_b
        L = self.Mx_b
        M = self.My_b
        N = self.Mz_b
        mass = self.mass_props.mass
        Ixx = self.mass_props.Ixx
        Iyy = self.mass_props.Iyy
        Izz = self.mass_props.Izz
        Ixy = self.mass_props.Ixy
        Iyz = self.mass_props.Iyz
        Ixz = self.mass_props.Ixz
        hx = self.hx_b
        hy = self.hy_b
        hz = self.hz_b

        ### Trig Shorthands
        def sincos(x):
            try:
                x = np.mod(x, 2 * np.pi)
                one = np.ones_like(x)
                zero = np.zeros_like(x)

                if np.allclose(x, 0) or np.allclose(x, 2 * np.pi):
                    sin = zero
                    cos = one
                elif np.allclose(x, np.pi / 2):
                    sin = one
                    cos = zero
                elif np.allclose(x, np.pi):
                    sin = zero
                    cos = -one
                elif np.allclose(x, 3 * np.pi / 2):
                    sin = -one
                    cos = zero
                else:
                    raise ValueError()
            except:
                sin = np.sin(x)
                cos = np.cos(x)
            return sin, cos

        # Do the trig
        sphi, cphi = sincos(phi)
        sthe, cthe = sincos(theta)
        spsi, cpsi = sincos(psi)

        ##### Equations of Motion

        ### Position derivatives
        d_xe = (
                (cthe * cpsi) * u +
                (sphi * sthe * cpsi - cphi * spsi) * v +
                (cphi * sthe * cpsi + sphi * spsi) * w
        )
        d_ye = (
                (cthe * spsi) * u +
                (sphi * sthe * spsi + cphi * cpsi) * v +
                (cphi * sthe * spsi - sphi * cpsi) * w
        )
        d_ze = (
                (-sthe) * u +
                (sphi * cthe) * v +
                (cphi * cthe) * w
        )
        ### Velocity derivatives
        d_u = (
                (X / mass) -
                q * w +
                r * v
        )
        d_v = (
                (Y / mass) -
                r * u +
                p * w
        )
        d_w = (
                (Z / mass) -
                p * v +
                q * u
        )
        ### Angle derivatives
        if np.all(cthe == 0):
            d_phi = 0
        else:
            d_phi = (
                    p +
                    q * sphi * sthe / cthe +
                    r * cphi * sthe / cthe
            )

        d_theta = (
                q * cphi -
                r * sphi
        )

        if np.all(cthe == 0):
            d_psi = 0
        else:
            d_psi = (
                    q * sphi / cthe +
                    r * cphi / cthe
            )

        ### Angular velocity derivatives
        RHS_L = (
                L -
                (Izz - Iyy) * q * r -
                Iyz * (q ** 2 - r ** 2) -
                Ixz * p * q +
                Ixy * p * r -
                hz * q +
                hy * r
        )
        RHS_M = (
                M -
                (Ixx - Izz) * r * p -
                Ixz * (r ** 2 - p ** 2) -
                Ixy * q * r +
                Iyz * q * p -
                hx * r +
                hz * p
        )
        RHS_N = (
                N -
                (Iyy - Ixx) * p * q -
                Ixy * (p ** 2 - q ** 2) -
                Iyz * r * p +
                Ixz * r * q -
                hy * p +
                hx * q
        )
        i11, i22, i33, i12, i23, i13 = np.linalg.inv_symmetric_3x3(Ixx, Iyy, Izz, Ixy, Iyz, Ixz)

        d_p = i11 * RHS_L + i12 * RHS_M + i13 * RHS_N
        d_q = i12 * RHS_L + i22 * RHS_M + i23 * RHS_N
        d_r = i13 * RHS_L + i23 * RHS_M + i33 * RHS_N

        return {
            "x_e"  : d_xe,
            "y_e"  : d_ye,
            "z_e"  : d_ze,
            "u_b"  : d_u,
            "v_b"  : d_v,
            "w_b"  : d_w,
            "phi"  : d_phi,
            "theta": d_theta,
            "psi"  : d_psi,
            "p"    : d_p,
            "q"    : d_q,
            "r"    : d_r,
        }

    def convert_axes(self,
                     x_from, y_from, z_from,
                     from_axes: str,
                     to_axes: str,
                     ):
        """
        Converts a vector [x_from, y_from, z_from], as given in the `from_axes` frame, to an equivalent vector [x_to,
        y_to, z_to], as given in the `to_axes` frame.

        Identical to OperatingPoint.convert_axes(), but adds in "earth" as a valid axis frame. For more documentation,
        see the docstring of OperatingPoint.convert_axes().

        Both `from_axes` and `to_axes` should be a string, one of:
                * "geometry"
                * "body"
                * "wind"
                * "stability"
                * "earth"

        Args:
                x_from: x-component of the vector, in `from_axes` frame.
                y_from: y-component of the vector, in `from_axes` frame.
                z_from: z-component of the vector, in `from_axes` frame.
                from_axes: The axes to convert from.
                to_axes: The axes to convert to.

        Returns: The x-, y-, and z-components of the vector, in `to_axes` frame. Given as a tuple.

        """
        if from_axes == to_axes:
            return x_from, y_from, z_from

        if from_axes == "earth" or to_axes == "earth":
            ### Trig Shorthands
            def sincos(x):
                try:
                    x = np.mod(x, 2 * np.pi)
                    one = np.ones_like(x)
                    zero = np.zeros_like(x)

                    if np.allclose(x, 0) or np.allclose(x, 2 * np.pi):
                        sin = zero
                        cos = one
                    elif np.allclose(x, np.pi / 2):
                        sin = one
                        cos = zero
                    elif np.allclose(x, np.pi):
                        sin = zero
                        cos = -one
                    elif np.allclose(x, 3 * np.pi / 2):
                        sin = -one
                        cos = zero
                    else:
                        raise ValueError()
                except:
                    sin = np.sin(x)
                    cos = np.cos(x)
                return sin, cos

                # Do the trig

            sphi, cphi = sincos(self.phi)
            sthe, cthe = sincos(self.theta)
            spsi, cpsi = sincos(self.psi)

        if from_axes == "earth":
            x_b = (
                    (cthe * cpsi) * x_from +
                    (cthe * spsi) * y_from +
                    (-sthe) * z_from
            )
            y_b = (
                    (sphi * sthe * cpsi - cphi * spsi) * x_from +
                    (sphi * sthe * spsi + cphi * cpsi) * y_from +
                    (sphi * cthe) * z_from
            )
            z_b = (
                    (cphi * sthe * cpsi + sphi * spsi) * x_from +
                    (cphi * sthe * spsi - sphi * cpsi) * y_from +
                    (cphi * cthe) * z_from
            )
        else:
            x_b, y_b, z_b = self.op_point.convert_axes(
                x_from, y_from, z_from,
                from_axes=from_axes, to_axes="body"
            )

        if to_axes == "earth":
            x_to = (
                    (cthe * cpsi) * x_b +
                    (sphi * sthe * cpsi - cphi * spsi) * y_b +
                    (cphi * sthe * cpsi + sphi * spsi) * z_b
            )
            y_to = (
                    (cthe * spsi) * x_b +
                    (sphi * sthe * spsi + cphi * cpsi) * y_b +
                    (cphi * sthe * spsi - sphi * cpsi) * z_b
            )
            z_to = (
                    (-sthe) * x_b +
                    (sphi * cthe) * y_b +
                    (cphi * cthe) * z_b
            )
        else:
            x_to, y_to, z_to = self.op_point.convert_axes(
                x_b, y_b, z_b,
                from_axes="body", to_axes=to_axes
            )

        return x_to, y_to, z_to

    def add_force(self,
                  Fx: Union[float, np.ndarray] = 0,
                  Fy: Union[float, np.ndarray] = 0,
                  Fz: Union[float, np.ndarray] = 0,
                  axes="body",
                  ):
        Fx_b, Fy_b, Fz_b = self.convert_axes(
            x_from=Fx,
            y_from=Fy,
            z_from=Fz,
            from_axes=axes,
            to_axes="body"
        )
        self.Fx_b = self.Fx_b + Fx_b
        self.Fy_b = self.Fy_b + Fy_b
        self.Fz_b = self.Fz_b + Fz_b

    def add_moment(self,
                   Mx: Union[float, np.ndarray] = 0,
                   My: Union[float, np.ndarray] = 0,
                   Mz: Union[float, np.ndarray] = 0,
                   axes="body",
                   ):
        Mx_b, My_b, Mz_b = self.convert_axes(
            x_from=Mx,
            y_from=My,
            z_from=Mz,
            from_axes=axes,
            to_axes="body"
        )
        self.Mx_b = self.Mx_b + Mx_b
        self.My_b = self.My_b + My_b
        self.Mz_b = self.Mz_b + Mz_b

    @property
    def speed(self):
        """The speed of the object, expressed as a scalar."""
        return (
                self.u_b ** 2 +
                self.v_b ** 2 +
                self.w_b ** 2
        ) ** 0.5

    @property
    def alpha(self):
        """The angle of attack, in degrees."""
        return np.arctan2d(
            self.w_b,
            self.u_b
        )

    @property
    def beta(self):
        """The sideslip angle, in degrees."""
        return np.arctan2d(
            self.v_b,
            (
                    self.u_b ** 2 +
                    self.w_b ** 2
            ) ** 0.5
        )


if __name__ == '__main__':
    import aerosandbox as asb

    dyn = DynamicsRigidBody3DBodyEuler(
        mass_props=asb.MassProperties(
            mass=1
        )
    )



import aerosandbox as asb
import aerosandbox.numpy as np

airfoil = asb.Airfoil("naca0008")

airplane = asb.Airplane(
    name="Flat Plate",
    xyz_ref=[0, 0, 0],
    wings=[
        asb.Wing(
            name="Wing",
            symmetric=True,
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=1,
                    twist=0,
                    airfoil=airfoil,
                ),
                asb.WingXSec(
                    xyz_le=[0, 5, 0],
                    chord=1,
                    twist=0,
                    airfoil=airfoil,
                ),
            ]
        )
    ]
)

if __name__ == '__main__':
    airplane.draw()



import aerosandbox as asb
import aerosandbox.numpy as np

af = asb.Airfoil("naca0012")
cache = asb._asb_root / "geometry/airfoil/test_airfoil/naca0012.json"


def make_cache():
    af.generate_polars(cache_filename=cache)


def test_load_cache():
    af.generate_polars(cache_filename=cache)


if __name__ == '__main__':
    make_cache()
    test_load_cache()



import aerosandbox.numpy as np


def peak_of_harmonic_oscillation(
        amplitude: float = 1,
        frequency_hz: float = 1,
        derivative_order: int = 0,
):
    """
    Computes the peak value of the nth derivative of a simple harmonic oscillation (e.g., sine wave).

    Specifically, if `t` represents time, and we have an oscillation that is described by a function f(t) as:

        `f(t) = amplitude * sin(2 * pi * t * frequency_hz)`

    then this function will return the value of the nth time derivative of f(t), where `derivative_order` gives the
    value of n.

    Example:

        My spring-mass system has an oscillating position, where the mass oscillates in position from -0.1 to 0.1,
        with a frequency of 3 Hz. I want to know what the peak acceleration on that mass is:

        >>> peak_of_harmonic_oscillation(
        >>>     amplitude=0.1,
        >>>     frequency_hz=3,
        >>>     derivative_order=2
        >>> )

        This would return (2 * pi * 3) ^ 2 * 0.1, or around 35.5.

        So, the peak acceleration on my mass is around 35.5 m/s^2.

    Args:

        amplitude: Amplitude of the underlying oscillation. Amplitude here is the center-to-peak distance,
        not peak-to-peak.

        frequency_hz: Frequency of the underlying oscillation, in units of 1/time (Hz). Note: not the angular frequency.

        derivative_order: The derivative order. (E.g., 0 for position, 1 for velocity, 2 for acceleration).

    Returns: The peak value of the nth derivative of a simple harmonic oscillator with the specified amplitude and frequency.

    """
    return (2 * np.pi * frequency_hz) ** derivative_order * amplitude



import aerosandbox as asb
import aerosandbox.numpy as np

wing_airfoil = asb.Airfoil("sd7037")
tail_airfoil = asb.Airfoil("naca0010")

### Define the 3D geometry you want to analyze/optimize.
# Here, all distances are in meters and all angles are in degrees.
airplane = asb.Airplane(
    name="Peter's Glider",
    xyz_ref=[0, 0, 0],  # CG location
    wings=[
        asb.Wing(
            name="Main Wing",
            symmetric=True,  # Should this wing be mirrored across the XZ plane?
            xsecs=[  # The wing's cross ("X") sections
                asb.WingXSec(  # Root
                    xyz_le=[0, 0, 0],  # Coordinates of the XSec's leading edge, relative to the wing's leading edge.
                    chord=0.18,
                    twist=2,  # degrees
                    airfoil=wing_airfoil,  # Airfoils are blended between a given XSec and the next one.
                ),
                asb.WingXSec(  # Mid
                    xyz_le=[0.01, 0.5, 0],
                    chord=0.16,
                    twist=0,
                    airfoil=wing_airfoil,
                ),
                asb.WingXSec(  # Tip
                    xyz_le=[0.08, 1, 0.1],
                    chord=0.08,
                    twist=-2,
                    airfoil=wing_airfoil,
                ),
            ]
        ),
        asb.Wing(
            name="Horizontal Stabilizer",
            symmetric=True,
            xsecs=[
                asb.WingXSec(  # root
                    xyz_le=[0, 0, 0],
                    chord=0.1,
                    twist=-10,
                    airfoil=tail_airfoil,
                ),
                asb.WingXSec(  # tip
                    xyz_le=[0.02, 0.17, 0],
                    chord=0.08,
                    twist=-10,
                    airfoil=tail_airfoil
                )
            ]
        ).translate([0.6, 0, 0.06]),
        asb.Wing(
            name="Vertical Stabilizer",
            symmetric=False,
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=0.1,
                    twist=0,
                    airfoil=tail_airfoil,
                ),
                asb.WingXSec(
                    xyz_le=[0.04, 0, 0.15],
                    chord=0.06,
                    twist=0,
                    airfoil=tail_airfoil
                )
            ]
        ).translate([0.6, 0, 0.07])
    ],
    fuselages=[
        asb.Fuselage(
            name="Fuselage",
            xsecs=[
                asb.FuselageXSec(
                    xyz_c=[0.8 * xi - 0.1, 0, 0.1 * xi - 0.03],
                    radius=0.6 * asb.Airfoil("dae51").local_thickness(x_over_c=xi)
                )
                for xi in np.cosspace(0, 1, 30)
            ]
        )
    ]
)

if __name__ == '__main__':
    airplane.draw()



from aerosandbox import *
import pytest


def w() -> Wing:
    wing = Wing(
        name="MyWing",
        xsecs=[
            WingXSec(
                xyz_le=np.array([1, 1, 0]),
                chord=0.5,
                twist=5,
                control_surfaces=[
                    ControlSurface(
                        symmetric=True,
                    )
                ]
            ),
            WingXSec(
                xyz_le=np.array([2, 2, 0]),
                chord=0.5,
                twist=5,
                control_surfaces=[
                    ControlSurface(
                        symmetric=True
                    )
                ]
            )
        ],
        symmetric=True
    ).translate(np.array([1, 2, 3]))
    return wing


def test_span():
    assert w().span() == pytest.approx(2)


def test_area():
    assert w().area() == pytest.approx(1)


def test_aspect_ratio():
    assert w().aspect_ratio() == pytest.approx(4)


def test_is_entirely_symmetric():
    assert w().is_entirely_symmetric()


def test_mean_geometric_chord():
    assert w().mean_geometric_chord() == pytest.approx(0.5)


def test_mean_aerodynamic_chord():
    assert w().mean_aerodynamic_chord() == pytest.approx(0.5)


def test_mean_twist_angle():
    assert w().mean_twist_angle() == pytest.approx(5)


def test_mean_sweep_angle():
    assert w().mean_sweep_angle() == pytest.approx(45)


def test_aerodynamic_center():
    ac = w().aerodynamic_center()

    assert ac[0] == pytest.approx(1 + 1.5 + 1 / 8, abs=2e-2)
    assert ac[1] == pytest.approx(0)
    assert ac[2] == pytest.approx(3, abs=2e-2)


if __name__ == '__main__':
    pytest.main()



import aerosandbox.numpy as np
from typing import Union, Tuple


def quadratic_bezier_patch_from_tangents(
        t: Union[float, np.ndarray],
        x_a: float,
        x_b: float,
        y_a: float,
        y_b: float,
        dydx_a: float,
        dydx_b: float,
) -> Tuple[Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Computes sampled points in 2D space from a quadratic Bezier spline defined by endpoints and end-tangents.

    Note: due to the inherent nature of a quadratic Bezier curve, curvature will be strictly one-sided - in other
    words, this will not make "S"-shaped curves. This means that you should be aware that bad values of dydx at
    either endpoint might cause this curvature to flip, which would result in the curve "going backwards" at one
    endpoint.

    Also, note that, in general, points will not be spaced evenly in x, y, or arc length s.

    Args:
        t:
        x_a: The x-coordinate of the first endpoint.
        x_b: The x-coordinate of the second endpoint.
        y_a: The y-coordinate of the first endpoint.
        y_b: The y-coordinate of the second endpoint.
        dydx_a: The derivative of y with respect to x at the first endpoint.
        dydx_b: The derivative of y with respect to x at the second endpoint.

    Returns:
        x: A scalar or numpy array of scalars representing the x-coordinates of the sampled points.
        y: A scalar or numpy array of scalars representing the y-coordinates of the sampled points.

    Usage:
        >>> x_a, x_b = 0, 10
        >>> y_a, y_b = 0, 5
        >>> dydx_a, dydx_b = 0.5, -0.5
        >>>
        >>> t = np.linspace(0, 1, 50)
        >>> x, y = quadratic_bezier_patch_from_tangents(
        >>>     t=t,
        >>>     x_a=x_a,
        >>>     x_b=x_b,
        >>>     y_a=y_a,
        >>>     y_b=y_b,
        >>>     dydx_a=dydx_a,
        >>>     dydx_b=dydx_b
        >>> )

    """
    ### Compute intercept of tangent lines
    x_P1 = (
                   (y_b - y_a) + (dydx_a * x_a - dydx_b * x_b)
           ) / (dydx_a - dydx_b)
    y_P1 = y_a + dydx_a * (x_P1 - x_a)

    x = (
            (1 - t) ** 2 * x_a +
            2 * (1 - t) * t * x_P1 +
            t ** 2 * x_b
    )
    y = (
            (1 - t) ** 2 * y_a +
            2 * (1 - t) * t * y_P1 +
            t ** 2 * y_b
    )

    return x, y

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    x, y = quadratic_bezier_patch_from_tangents(
        t = np.linspace(0, 1, 11),
        x_a=1,
        x_b=4,
        y_a=2,
        y_b=3,
        dydx_a=1,
        dydx_b=-30
    )

    plt.plot(x, y, ".-")
    p.show_plot()


import matplotlib
import mpl_toolkits
import matplotlib.pyplot as plt
import aerosandbox.numpy as np
from typing import Dict, Tuple, Union

preset_view_angles = {
    # Given in the form:
    #   * key is the view name
    #   * value is a tuple of three floats: (elev, azim, roll)
    'XY'             : (90, -90, 0),
    'XZ'             : (0, -90, 0),
    'YZ'             : (0, 0, 0),
    '-XY'            : (-90, 90, 0),
    '-XZ'            : (0, 90, 0),
    '-YZ'            : (0, 180, 0),
    'left_isometric' : (np.arctan2d(1, 2 ** 0.5), -135, 0),
    'right_isometric': (np.arctan2d(1, 2 ** 0.5), 135, 0)
}
preset_view_angles['front'] = preset_view_angles["-YZ"]
preset_view_angles['top'] = preset_view_angles["XY"]
preset_view_angles['side'] = preset_view_angles["XZ"]


def figure3d(
        nrows: int = 1,
        ncols: int = 1,
        orthographic: bool = True,
        box_aspect: Tuple[float] = None,
        adjust_colors: bool = True,
        ax_kwargs: Dict = None,
        **fig_kwargs
) -> Tuple[matplotlib.figure.Figure, mpl_toolkits.mplot3d.axes3d.Axes3D]:
    """
    Creates a new 3D figure. Args and kwargs are passed into matplotlib.pyplot.figure().

    Returns: (fig, ax)

    """
    ### Set defaults
    if ax_kwargs is None:
        ax_kwargs = {}

    ### Collect the keyword arguments to be used for each 3D axis
    default_axes_kwargs = dict(
        projection='3d',
        proj_type='ortho' if orthographic else 'persp',
        box_aspect=box_aspect,
    )
    axes_kwargs = {  # Overwrite any of the computed kwargs with user-provided ones, where applicable.
        **default_axes_kwargs,
        **ax_kwargs,
    }

    ### Generate the 3D axis (or axes)
    fig, ax = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        subplot_kw=axes_kwargs,
        **fig_kwargs
    )

    if adjust_colors:
        try:
            axs = ax.flatten()
        except AttributeError:
            axs = [ax]

        for a in axs:

            pane_color = a.get_facecolor()
            a.set_facecolor((0, 0, 0, 0))  # Set transparent

            a.xaxis.pane.set_facecolor(pane_color)
            a.xaxis.pane.set_alpha(1)
            a.yaxis.pane.set_facecolor(pane_color)
            a.yaxis.pane.set_alpha(1)
            a.zaxis.pane.set_facecolor(pane_color)
            a.zaxis.pane.set_alpha(1)

    return fig, ax


def ax_is_3d(
        ax: matplotlib.axes.Axes = None
) -> bool:
    """
    Determines if a Matplotlib axis object is 3D or not.

    Args:
        ax: The axis object. If not given, uses the current active axes.

    Returns: A boolean of whether the axis is 3D or not.

    """
    if ax is None:
        ax = plt.gca()

    return hasattr(ax, 'zaxis')


def set_preset_3d_view_angle(
        preset_view: str
) -> None:
    ax = plt.gca()

    if not ax_is_3d(ax):
        raise Exception("Can't set a 3D view angle on a non-3D plot!")

    try:
        elev, azim, roll = preset_view_angles[preset_view]
    except KeyError:
        raise ValueError(
            f"Input '{preset_view}' is not a valid preset view. Valid presets are:\n" +
            "\n".join([f"  * '{k}'" for k in preset_view_angles.keys()])
        )

    if roll == 0:
        # This is to maintain back-compatibility to older Matplotlib versions.
        # Older versions of Matplotlib (roughly, <=3.4.0) didn't support the `roll` kwarg.
        # Hence, if we don't need to edit the roll, we don't - this extends back-compatibility.
        ax.view_init(
            elev=elev,
            azim=azim,
        )
    else:
        ax.view_init(
            elev=elev,
            azim=azim,
            roll=roll
        )


if __name__ == '__main__':
    import aerosandbox.numpy as np
    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    t = np.linspace(0, 1, 100)

    x = np.sin(4 * 2 * np.pi * t)
    y = t ** 2
    z = 5 * t

    fig, ax = p.figure3d()
    p.set_preset_3d_view_angle('left_isometric')

    ax.plot(
        x, y, z, "-"
    )
    p.set_ticks(*([1, 1] * 3))
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")

    p.equal()
    p.show_plot()



import numpy as np
from aerosandbox import ExplicitAnalysis
from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
from aerosandbox.aerodynamics.aero_3D.singularities.uniform_strength_horseshoe_singularities import \
    calculate_induced_velocity_horseshoe
from typing import Dict, Any, List, Callable
import copy


### Define some helper functions that take a vector and make it a Nx1 or 1xN, respectively.
# Useful for broadcasting with matrices later.
def tall(array):
    return np.reshape(array, (-1, 1))


def wide(array):
    return np.reshape(array, (1, -1))


class VortexLatticeMethod(ExplicitAnalysis):
    """
    An explicit (linear) vortex-lattice-method aerodynamics analysis.

    Usage example:
        >>> analysis = asb.VortexLatticeMethod(
        >>>     airplane=my_airplane,
        >>>     op_point=asb.OperatingPoint(
        >>>         velocity=100, # m/s
        >>>         alpha=5, # deg
        >>>         beta=4, # deg
        >>>         p=0.01, # rad/sec
        >>>         q=0.02, # rad/sec
        >>>         r=0.03, # rad/sec
        >>>     )
        >>> )
        >>> aero_data = analysis.run()
        >>> analysis.draw()
    """

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 xyz_ref: List[float] = None,
                 run_symmetric_if_possible: bool = False,
                 verbose: bool = False,
                 spanwise_resolution: int = 10,
                 spanwise_spacing_function: Callable[[float, float, float], np.ndarray] = np.cosspace,
                 chordwise_resolution: int = 10,
                 chordwise_spacing_function: Callable[[float, float, float], np.ndarray] = np.cosspace,
                 vortex_core_radius: float = 1e-8,
                 align_trailing_vortices_with_wind: bool = False,
                 ):
        super().__init__()

        ### Set defaults
        if xyz_ref is None:
            xyz_ref = airplane.xyz_ref

        self.airplane = airplane
        self.op_point = op_point
        self.xyz_ref = xyz_ref
        self.verbose = verbose
        self.spanwise_resolution = spanwise_resolution
        self.spanwise_spacing_function = spanwise_spacing_function
        self.chordwise_resolution = chordwise_resolution
        self.chordwise_spacing_function = chordwise_spacing_function
        self.vortex_core_radius = vortex_core_radius
        self.align_trailing_vortices_with_wind = align_trailing_vortices_with_wind

        ### Determine whether you should run the problem as symmetric
        self.run_symmetric = False
        if run_symmetric_if_possible:
            raise NotImplementedError("VLM with symmetry detection not yet implemented!")
            # try:
            #     self.run_symmetric = (  # Satisfies assumptions
            #             self.op_point.beta == 0 and
            #             self.op_point.p == 0 and
            #             self.op_point.r == 0 and
            #             self.airplane.is_entirely_symmetric()
            #     )
            # except RuntimeError:  # Required because beta, p, r, etc. may be non-numeric (e.g. opti variables)
            #     pass

    def run(self) -> Dict[str, Any]:
        """
        Computes the aerodynamic forces.

        Returns a dictionary with keys:

            - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
            - 'F_b' : an [x, y, z] list of forces in body axes [N]
            - 'F_w' : an [x, y, z] list of forces in wind axes [N]
            - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
            - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
            - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
            - 'L' : the lift force [N]. Definitionally, this is in wind axes.
            - 'Y' : the side force [N]. This is in wind axes.
            - 'D' : the drag force [N]. Definitionally, this is in wind axes.
            - 'l_b', the rolling moment, in body axes [Nm]. Positive is roll-right.
            - 'm_b', the pitching moment, in body axes [Nm]. Positive is pitch-up.
            - 'n_b', the yawing moment, in body axes [Nm]. Positive is nose-right.
            - 'CL', the lift coefficient [-]. Definitionally, this is in wind axes.
            - 'CY', the sideforce coefficient [-]. This is in wind axes.
            - 'CD', the drag coefficient [-]. Definitionally, this is in wind axes.
            - 'Cl', the rolling coefficient [-], in body axes
            - 'Cm', the pitching coefficient [-], in body axes
            - 'Cn', the yawing coefficient [-], in body axes

        Nondimensional values are nondimensionalized using reference values in the VortexLatticeMethod.airplane object.
        """

        if self.verbose:
            print("Meshing...")

        ##### Make Panels
        front_left_vertices = []
        back_left_vertices = []
        back_right_vertices = []
        front_right_vertices = []
        is_trailing_edge = []

        for wing in self.airplane.wings:
            if self.spanwise_resolution > 1:
                wing = wing.subdivide_sections(
                    ratio=self.spanwise_resolution,
                    spacing_function=self.spanwise_spacing_function
                )

            points, faces = wing.mesh_thin_surface(
                method="quad",
                chordwise_resolution=self.chordwise_resolution,
                chordwise_spacing_function=self.chordwise_spacing_function,
                add_camber=True
            )
            front_left_vertices.append(points[faces[:, 0], :])
            back_left_vertices.append(points[faces[:, 1], :])
            back_right_vertices.append(points[faces[:, 2], :])
            front_right_vertices.append(points[faces[:, 3], :])
            is_trailing_edge.append(
                (np.arange(len(faces)) + 1) % self.chordwise_resolution == 0
            )

        front_left_vertices = np.concatenate(front_left_vertices)
        back_left_vertices = np.concatenate(back_left_vertices)
        back_right_vertices = np.concatenate(back_right_vertices)
        front_right_vertices = np.concatenate(front_right_vertices)
        is_trailing_edge = np.concatenate(is_trailing_edge)

        ### Compute panel statistics
        diag1 = front_right_vertices - back_left_vertices
        diag2 = front_left_vertices - back_right_vertices
        cross = np.cross(diag1, diag2)
        cross_norm = np.linalg.norm(cross, axis=1)
        normal_directions = cross / tall(cross_norm)
        areas = cross_norm / 2

        # Compute the location of points of interest on each panel
        left_vortex_vertices = 0.75 * front_left_vertices + 0.25 * back_left_vertices
        right_vortex_vertices = 0.75 * front_right_vertices + 0.25 * back_right_vertices
        vortex_centers = (left_vortex_vertices + right_vortex_vertices) / 2
        vortex_bound_leg = right_vortex_vertices - left_vortex_vertices
        collocation_points = (
                0.5 * (0.25 * front_left_vertices + 0.75 * back_left_vertices) +
                0.5 * (0.25 * front_right_vertices + 0.75 * back_right_vertices)
        )

        ### Save things to the instance for later access
        self.front_left_vertices = front_left_vertices
        self.back_left_vertices = back_left_vertices
        self.back_right_vertices = back_right_vertices
        self.front_right_vertices = front_right_vertices
        self.is_trailing_edge = is_trailing_edge
        self.normal_directions = normal_directions
        self.areas = areas
        self.left_vortex_vertices = left_vortex_vertices
        self.right_vortex_vertices = right_vortex_vertices
        self.vortex_centers = vortex_centers
        self.vortex_bound_leg = vortex_bound_leg
        self.collocation_points = collocation_points

        ##### Setup Operating Point
        if self.verbose:
            print("Calculating the freestream influence...")
        steady_freestream_velocity = self.op_point.compute_freestream_velocity_geometry_axes()  # Direction the wind is GOING TO, in geometry axes coordinates
        steady_freestream_direction = steady_freestream_velocity / np.linalg.norm(steady_freestream_velocity)
        rotation_freestream_velocities = self.op_point.compute_rotation_velocity_geometry_axes(
            collocation_points)

        freestream_velocities = np.add(wide(steady_freestream_velocity), rotation_freestream_velocities)
        # Nx3, represents the freestream velocity at each panel collocation point (c)

        freestream_influences = np.sum(freestream_velocities * normal_directions, axis=1)

        ### Save things to the instance for later access
        self.steady_freestream_velocity = steady_freestream_velocity
        self.steady_freestream_direction = steady_freestream_direction
        self.freestream_velocities = freestream_velocities

        ##### Setup Geometry
        ### Calculate AIC matrix
        if self.verbose:
            print("Calculating the collocation influence matrix...")

        u_collocations_unit, v_collocations_unit, w_collocations_unit = calculate_induced_velocity_horseshoe(
            x_field=tall(collocation_points[:, 0]),
            y_field=tall(collocation_points[:, 1]),
            z_field=tall(collocation_points[:, 2]),
            x_left=wide(left_vortex_vertices[:, 0]),
            y_left=wide(left_vortex_vertices[:, 1]),
            z_left=wide(left_vortex_vertices[:, 2]),
            x_right=wide(right_vortex_vertices[:, 0]),
            y_right=wide(right_vortex_vertices[:, 1]),
            z_right=wide(right_vortex_vertices[:, 2]),
            trailing_vortex_direction=(
                steady_freestream_direction
                if self.align_trailing_vortices_with_wind else
                np.array([1, 0, 0])
            ),
            gamma=1.,
            vortex_core_radius=self.vortex_core_radius
        )

        AIC = (
                u_collocations_unit * tall(normal_directions[:, 0]) +
                v_collocations_unit * tall(normal_directions[:, 1]) +
                w_collocations_unit * tall(normal_directions[:, 2])
        )

        ##### Calculate Vortex Strengths
        if self.verbose:
            print("Calculating vortex strengths...")

        self.vortex_strengths = np.linalg.solve(AIC, -freestream_influences)

        ##### Calculate forces
        ### Calculate Near-Field Forces and Moments
        # Governing Equation: The force on a straight, small vortex filament is F = rho * cross(V, l) * gamma,
        # where rho is density, V is the velocity vector, cross() is the cross product operator,
        # l is the vector of the filament itself, and gamma is the circulation.

        if self.verbose:
            print("Calculating forces on each panel...")
        # Calculate the induced velocity at the center of each bound leg
        V_centers = self.get_velocity_at_points(vortex_centers)

        # Calculate forces_inviscid_geometry, the force on the ith panel. Note that this is in GEOMETRY AXES,
        # not WIND AXES or BODY AXES.
        Vi_cross_li = np.cross(V_centers, vortex_bound_leg, axis=1)

        forces_geometry = self.op_point.atmosphere.density() * Vi_cross_li * tall(self.vortex_strengths)
        moments_geometry = np.cross(
            np.add(vortex_centers, -wide(np.array(self.xyz_ref))),
            forces_geometry
        )

        # Calculate total forces and moments
        force_geometry = np.sum(forces_geometry, axis=0)
        moment_geometry = np.sum(moments_geometry, axis=0)

        force_body = self.op_point.convert_axes(
            force_geometry[0], force_geometry[1], force_geometry[2],
            from_axes="geometry",
            to_axes="body"
        )
        force_wind = self.op_point.convert_axes(
            force_body[0], force_body[1], force_body[2],
            from_axes="body",
            to_axes="wind"
        )
        moment_body = self.op_point.convert_axes(
            moment_geometry[0], moment_geometry[1], moment_geometry[2],
            from_axes="geometry",
            to_axes="body"
        )
        moment_wind = self.op_point.convert_axes(
            moment_body[0], moment_body[1], moment_body[2],
            from_axes="body",
            to_axes="wind"
        )

        ### Save things to the instance for later access
        self.forces_geometry = forces_geometry
        self.moments_geometry = moments_geometry
        self.force_geometry = force_geometry
        self.force_body = force_body
        self.force_wind = force_wind
        self.moment_geometry = moment_geometry
        self.moment_body = moment_body
        self.moment_wind = moment_wind

        # Calculate dimensional forces
        L = -force_wind[2]
        D = -force_wind[0]
        Y = force_wind[1]
        l_b = moment_body[0]
        m_b = moment_body[1]
        n_b = moment_body[2]

        # Calculate nondimensional forces
        q = self.op_point.dynamic_pressure()
        s_ref = self.airplane.s_ref
        b_ref = self.airplane.b_ref
        c_ref = self.airplane.c_ref
        CL = L / q / s_ref
        CD = D / q / s_ref
        CY = Y / q / s_ref
        Cl = l_b / q / s_ref / b_ref
        Cm = m_b / q / s_ref / c_ref
        Cn = n_b / q / s_ref / b_ref

        return {
            "F_g": force_geometry,
            "F_b": force_body,
            "F_w": force_wind,
            "M_g": moment_geometry,
            "M_b": moment_body,
            "M_w": moment_wind,
            "L"  : L,
            "D"  : D,
            "Y"  : Y,
            "l_b": l_b,
            "m_b": m_b,
            "n_b": n_b,
            "CL" : CL,
            "CD" : CD,
            "CY" : CY,
            "Cl" : Cl,
            "Cm" : Cm,
            "Cn" : Cn,
        }

    def run_with_stability_derivatives(self,
                                       alpha=True,
                                       beta=True,
                                       p=True,
                                       q=True,
                                       r=True,
                                       ):
        """
                Computes the aerodynamic forces and moments on the airplane, and the stability derivatives.

                Arguments essentially determine which stability derivatives are computed. If a stability derivative is not
                needed, leaving it False will speed up the computation.

                Args:

                    - alpha (bool): If True, compute the stability derivatives with respect to the angle of attack (alpha).
                    - beta (bool): If True, compute the stability derivatives with respect to the sideslip angle (beta).
                    - p (bool): If True, compute the stability derivatives with respect to the body-axis roll rate (p).
                    - q (bool): If True, compute the stability derivatives with respect to the body-axis pitch rate (q).
                    - r (bool): If True, compute the stability derivatives with respect to the body-axis yaw rate (r).

                Returns: a dictionary with keys:

                    - 'F_g' : an [x, y, z] list of forces in geometry axes [N]
                    - 'F_b' : an [x, y, z] list of forces in body axes [N]
                    - 'F_w' : an [x, y, z] list of forces in wind axes [N]
                    - 'M_g' : an [x, y, z] list of moments about geometry axes [Nm]
                    - 'M_b' : an [x, y, z] list of moments about body axes [Nm]
                    - 'M_w' : an [x, y, z] list of moments about wind axes [Nm]
                    - 'L' : the lift force [N]. Definitionally, this is in wind axes.
                    - 'Y' : the side force [N]. This is in wind axes.
                    - 'D' : the drag force [N]. Definitionally, this is in wind axes.
                    - 'l_b', the rolling moment, in body axes [Nm]. Positive is roll-right.
                    - 'm_b', the pitching moment, in body axes [Nm]. Positive is pitch-up.
                    - 'n_b', the yawing moment, in body axes [Nm]. Positive is nose-right.
                    - 'CL', the lift coefficient [-]. Definitionally, this is in wind axes.
                    - 'CY', the sideforce coefficient [-]. This is in wind axes.
                    - 'CD', the drag coefficient [-]. Definitionally, this is in wind axes.
                    - 'Cl', the rolling coefficient [-], in body axes
                    - 'Cm', the pitching coefficient [-], in body axes
                    - 'Cn', the yawing coefficient [-], in body axes

                    Along with additional keys, depending on the value of the `alpha`, `beta`, `p`, `q`, and `r` arguments. For
                    example, if `alpha=True`, then the following additional keys will be present:

                        - 'CLa', the lift coefficient derivative with respect to alpha [1/rad]
                        - 'CDa', the drag coefficient derivative with respect to alpha [1/rad]
                        - 'CYa', the sideforce coefficient derivative with respect to alpha [1/rad]
                        - 'Cla', the rolling moment coefficient derivative with respect to alpha [1/rad]
                        - 'Cma', the pitching moment coefficient derivative with respect to alpha [1/rad]
                        - 'Cna', the yawing moment coefficient derivative with respect to alpha [1/rad]
                        - 'x_np', the neutral point location in the x direction [m]

                    Nondimensional values are nondimensionalized using reference values in the
                    VortexLatticeMethod.airplane object.

                    Data types:
                        - The "L", "Y", "D", "l_b", "m_b", "n_b", "CL", "CY", "CD", "Cl", "Cm", and "Cn" keys are:

                            - floats if the OperatingPoint object is not vectorized (i.e., if all attributes of OperatingPoint
                            are floats, not arrays).

                            - arrays if the OperatingPoint object is vectorized (i.e., if any attribute of OperatingPoint is an
                            array).

                        - The "F_g", "F_b", "F_w", "M_g", "M_b", and "M_w" keys are always lists, which will contain either
                        floats or arrays, again depending on whether the OperatingPoint object is vectorized or not.

                """
        abbreviations = {
            "alpha": "a",
            "beta" : "b",
            "p"    : "p",
            "q"    : "q",
            "r"    : "r",
        }
        finite_difference_amounts = {
            "alpha": 0.001,
            "beta" : 0.001,
            "p"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.b_ref,
            "q"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.c_ref,
            "r"    : 0.001 * (2 * self.op_point.velocity) / self.airplane.b_ref,
        }
        scaling_factors = {
            "alpha": np.degrees(1),
            "beta" : np.degrees(1),
            "p"    : (2 * self.op_point.velocity) / self.airplane.b_ref,
            "q"    : (2 * self.op_point.velocity) / self.airplane.c_ref,
            "r"    : (2 * self.op_point.velocity) / self.airplane.b_ref,
        }

        original_op_point = self.op_point

        # Compute the point analysis, which returns a dictionary that we will later add key:value pairs to.
        run_base = self.run()

        # Note for the loops below: here, "derivative numerator" and "... denominator" refer to the quantity being
        # differentiated and the variable of differentiation, respectively. In other words, in the expression df/dx,
        # the "numerator" is f, and the "denominator" is x. I realize that this would make a mathematician cry (as a
        # partial derivative is not a fraction), but the reality is that there seems to be no commonly-accepted name
        # for these terms. (Curiously, this contrasts with integration, where there is an "integrand" and a "variable
        # of integration".)

        for derivative_denominator in abbreviations.keys():
            if not locals()[derivative_denominator]:  # Basically, if the parameter from the function input is not True,
                continue  # Skip this run.
                # This way, you can (optionally) speed up this routine if you only need static derivatives,
                # or longitudinal derivatives, etc.

            # These lines make a copy of the original operating point, incremented by the finite difference amount
            # along the variable defined by derivative_denominator.
            incremented_op_point = copy.copy(original_op_point)
            incremented_op_point.__setattr__(
                derivative_denominator,
                original_op_point.__getattribute__(derivative_denominator) + finite_difference_amounts[
                    derivative_denominator]
            )

            vlm_incremented = copy.copy(self)
            vlm_incremented.op_point = incremented_op_point
            run_incremented = vlm_incremented.run()

            for derivative_numerator in [
                "CL",
                "CD",
                "CY",
                "Cl",
                "Cm",
                "Cn",
            ]:
                derivative_name = derivative_numerator + abbreviations[derivative_denominator]  # Gives "CLa"
                run_base[derivative_name] = (
                        (  # Finite-difference out the derivatives
                                run_incremented[derivative_numerator] - run_base[
                            derivative_numerator]
                        ) / finite_difference_amounts[derivative_denominator]
                        * scaling_factors[derivative_denominator]
                )

            ### Try to compute and append neutral point, if possible
            if derivative_denominator == "alpha":
                run_base["x_np"] = self.xyz_ref[0] - (
                        run_base["Cma"] * (self.airplane.c_ref / run_base["CLa"])
                )
            if derivative_denominator == "beta":
                run_base["x_np_lateral"] = self.xyz_ref[0] - (
                        run_base["Cnb"] * (self.airplane.b_ref / run_base["CYb"])
                )

        return run_base

    def get_induced_velocity_at_points(self, points: np.ndarray) -> np.ndarray:
        """
        Computes the induced velocity at a set of points in the flowfield.

        Args:
            points: A Nx3 array of points that you would like to know the induced velocities at. Given in geometry axes.

        Returns: A Nx3 of the induced velocity at those points. Given in geometry axes.

        """
        u_induced, v_induced, w_induced = calculate_induced_velocity_horseshoe(
            x_field=tall(points[:, 0]),
            y_field=tall(points[:, 1]),
            z_field=tall(points[:, 2]),
            x_left=wide(self.left_vortex_vertices[:, 0]),
            y_left=wide(self.left_vortex_vertices[:, 1]),
            z_left=wide(self.left_vortex_vertices[:, 2]),
            x_right=wide(self.right_vortex_vertices[:, 0]),
            y_right=wide(self.right_vortex_vertices[:, 1]),
            z_right=wide(self.right_vortex_vertices[:, 2]),
            trailing_vortex_direction=self.steady_freestream_direction if self.align_trailing_vortices_with_wind else np.array(
                [1, 0, 0]),
            gamma=wide(self.vortex_strengths),
            vortex_core_radius=self.vortex_core_radius
        )
        u_induced = np.sum(u_induced, axis=1)
        v_induced = np.sum(v_induced, axis=1)
        w_induced = np.sum(w_induced, axis=1)

        V_induced = np.stack([
            u_induced, v_induced, w_induced
        ], axis=1)

        return V_induced

    def get_velocity_at_points(self, points: np.ndarray) -> np.ndarray:
        """
        Computes the velocity at a set of points in the flowfield.

        Args:
            points: A Nx3 array of points that you would like to know the velocities at. Given in geometry axes.

        Returns: A Nx3 of the velocity at those points. Given in geometry axes.

        """
        V_induced = self.get_induced_velocity_at_points(points)

        rotation_freestream_velocities = self.op_point.compute_rotation_velocity_geometry_axes(
            points
        )

        freestream_velocities = np.add(wide(self.steady_freestream_velocity), rotation_freestream_velocities)

        V = V_induced + freestream_velocities
        return V

    def calculate_streamlines(self,
                              seed_points: np.ndarray = None,
                              n_steps: int = 300,
                              length: float = None,
                              ) -> np.ndarray:
        """
        Computes streamlines, starting at specific seed points.

        After running this function, a new instance variable `VortexLatticeFilaments.streamlines` is computed

        Uses simple forward-Euler integration with a fixed spatial stepsize (i.e., velocity vectors are normalized
        before ODE integration). After investigation, it's not worth doing fancier ODE integration methods (adaptive
        schemes, RK substepping, etc.), due to the near-singular conditions near vortex filaments.

        Args:

            seed_points: A Nx3 ndarray that contains a list of points where streamlines are started. Will be
            auto-calculated if not specified.

            n_steps: The number of individual streamline steps to trace. Minimum of 2.

            length: The approximate total length of the streamlines desired, in meters. Will be auto-calculated if
            not specified.

        Returns:
            streamlines: a 3D array with dimensions: (n_seed_points) x (3) x (n_steps).
            Consists of streamlines data.

            Result is also saved as an instance variable, VortexLatticeMethod.streamlines.

        """
        if self.verbose:
            print("Calculating streamlines...")
        if length is None:
            length = self.airplane.c_ref * 5
        if seed_points is None:
            left_TE_vertices = self.back_left_vertices[self.is_trailing_edge.astype(bool)]
            right_TE_vertices = self.back_right_vertices[self.is_trailing_edge.astype(bool)]
            N_streamlines_target = 200
            seed_points_per_panel = np.maximum(1, N_streamlines_target // len(left_TE_vertices))

            nondim_node_locations = np.linspace(0, 1, seed_points_per_panel + 1)
            nondim_seed_locations = (nondim_node_locations[1:] + nondim_node_locations[:-1]) / 2

            seed_points = np.concatenate([
                x * left_TE_vertices + (1 - x) * right_TE_vertices
                for x in nondim_seed_locations
            ])

        streamlines = np.empty((len(seed_points), 3, n_steps))
        streamlines[:, :, 0] = seed_points
        for i in range(1, n_steps):
            V = self.get_velocity_at_points(streamlines[:, :, i - 1])
            streamlines[:, :, i] = (
                    streamlines[:, :, i - 1] +
                    length / n_steps * V / tall(np.linalg.norm(V, axis=1))
            )

        self.streamlines = streamlines

        if self.verbose:
            print("Streamlines calculated.")

        return streamlines

    def draw(self,
             c: np.ndarray = None,
             cmap: str = None,
             colorbar_label: str = None,
             show: bool = True,
             show_kwargs: Dict = None,
             draw_streamlines=True,
             recalculate_streamlines=False,
             backend: str = "pyvista"
             ):
        """
        Draws the solution. Note: Must be called on a SOLVED AeroProblem object.
        To solve an AeroProblem, use opti.solve(). To substitute a solved solution, use ap = ap.substitute_solution(sol).
        :return:
        """
        if show_kwargs is None:
            show_kwargs = {}

        if c is None:
            c = self.vortex_strengths
            colorbar_label = "Vortex Strengths"

        if draw_streamlines:
            if (not hasattr(self, 'streamlines')) or recalculate_streamlines:
                self.calculate_streamlines()

        if backend == "plotly":
            from aerosandbox.visualization.plotly_Figure3D import Figure3D
            fig = Figure3D()

            for i in range(len(self.front_left_vertices)):
                fig.add_quad(
                    points=[
                        self.front_left_vertices[i, :],
                        self.back_left_vertices[i, :],
                        self.back_right_vertices[i, :],
                        self.front_right_vertices[i, :],
                    ],
                    intensity=c[i],
                    outline=True,
                )

            if draw_streamlines:
                for i in range(self.streamlines.shape[0]):
                    fig.add_streamline(self.streamlines[i, :, :].T)

            return fig.draw(
                show=show,
                colorbar_title=colorbar_label,
                **show_kwargs,
            )

        elif backend == "pyvista":
            import pyvista as pv
            plotter = pv.Plotter()
            plotter.title = "ASB VortexLatticeMethod"
            plotter.add_axes()
            plotter.show_grid(color='gray')

            ### Draw the airplane mesh
            points = np.concatenate([
                self.front_left_vertices,
                self.back_left_vertices,
                self.back_right_vertices,
                self.front_right_vertices
            ])
            N = len(self.front_left_vertices)
            range_N = np.arange(N)
            faces = tall(range_N) + wide(np.array([0, 1, 2, 3]) * N)

            mesh = pv.PolyData(
                *mesh_utils.convert_mesh_to_polydata_format(points, faces)
            )
            scalar_bar_args = {}
            if colorbar_label is not None:
                scalar_bar_args["title"] = colorbar_label
            plotter.add_mesh(
                mesh=mesh,
                scalars=c,
                show_edges=True,
                show_scalar_bar=c is not None,
                scalar_bar_args=scalar_bar_args,
                cmap=cmap,
            )

            ### Draw the streamlines
            if draw_streamlines:
                import aerosandbox.tools.pretty_plots as p
                for i in range(self.streamlines.shape[0]):
                    plotter.add_mesh(
                        pv.Spline(self.streamlines[i, :, :].T),
                        color=p.adjust_lightness("#7700FF", 1.5),
                        opacity=0.7,
                        line_width=1
                    )

            if show:
                plotter.show(**show_kwargs)
            return plotter

        else:
            raise ValueError("Bad value of `backend`!")


if __name__ == '__main__':
    ### Import Vanilla Airplane
    import aerosandbox as asb

    from pathlib import Path

    geometry_folder = Path(__file__).parent / "test_aero_3D" / "geometries"

    import sys

    sys.path.insert(0, str(geometry_folder))

    from vanilla import airplane as vanilla

    ### Do the AVL run
    vlm = VortexLatticeMethod(
        airplane=vanilla,
        op_point=asb.OperatingPoint(
            atmosphere=asb.Atmosphere(altitude=0),
            velocity=10,
            alpha=0,
            beta=0,
            p=0,
            q=0,
            r=0,
        ),
        spanwise_resolution=12,
        chordwise_resolution=12,
    )

    res = vlm.run()

    for k, v in res.items():
        print(f"{str(k).rjust(10)} : {v}")



from aerosandbox.tools.inspect_tools import *
import pytest
import inspect


def test_function_argument_names_from_source_code():
    tests = {  # Pairs of {input: expected_output}
        "f(a, b)"               : ['a', 'b'],
        "f(a,b)"                : ['a', 'b'],
        "f(\na,\nb)"            : ['a', 'b'],
        "g = f(a, b)"           : ['a', 'b'],
        "g.h = f(a, b)"         : ['a', 'b'],
        "g.h() = f(a, b)"       : ['a', 'b'],
        "g.h(i=j) = f(a, b)"    : ['a', 'b'],
        "f(a, b) + g(h)"        : ['a', 'b'],
        "f(a: int, b: MyType())": ['a', 'b'],
        "f(a, b).g(c, d)"       : ['a', 'b'],
        "f(a(b), c)"            : ['a(b)', 'c'],
        "f(a(b,c), d)"          : ['a(b,c)', 'd'],
        "f({a:b}, c)"           : ['{a:b}', 'c'],
        "f(a[b], c)"            : ['a[b]', 'c'],
        "f({a:b, c:d}, e)"      : ['{a:b, c:d}', 'e'],
        "f({a:b,\nc:d}, e)"     : ['{a:b,c:d}', 'e'],
        "f(dict(a=b,c=d), e)"   : ['dict(a=b,c=d)', 'e'],
        "f(a=1, b=2)"           : ['a=1', 'b=2'],
        "f()"                   : [''],
        "f(a, [i for i in l])"  : ['a', '[i for i in l]'],
        "f(incomplete, "        : ValueError,
        "3 + 5"                 : ValueError,
        ""                      : ValueError,
    }

    for input, expected_output in tests.items():

        ### If you're expecting an error, make sure it gets raised
        if inspect.isclass(expected_output) and issubclass(expected_output, Exception):
            with pytest.raises(expected_output):
                get_function_argument_names_from_source_code(input)

        ### If you're expecting a specific output, make sure you get that
        else:
            assert get_function_argument_names_from_source_code(input) == expected_output


def test_codegen_builtins():
    for x in [
        1,
        1.0,
        "hello",
        True,
        None,
        [1, 2, 3],
        (1, 2, 3),
        {1, 2, 3},
        {1: 2, 3: 4},
        dict(cat=1, dog=2),
        [1, 2, [3, 4, [5, 6]]],
    ]:

        code, imports = codegen(x)
        for import_str in imports:
            exec(import_str)
        assert eval(code) == x


def test_codegen_numpy():
    import numpy as _np  # doing this to hide the import from the codegen function

    for x in [
        _np.array([1, 2, 3]),
        _np.arange(10),
        _np.arange(12).reshape(3, 4),
    ]:

        code, imports = codegen(x)
        for import_str in imports:
            exec(import_str)
        assert (eval(code) == x).all()


if __name__ == '__main__':
    test_function_argument_names_from_source_code()
    pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np

import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p

fuselage = asb.Fuselage(
    xsecs=[
        asb.FuselageXSec(
            xyz_c=[xi, 0, 0],
            radius=asb.Airfoil("naca0020").local_thickness(xi) / 2
        )
        for xi in np.cosspace(0, 1, 20)
    ],
)


# fuselage.draw()


def get_aero(xyz_ref):
    return asb.AeroBuildup(
        airplane=asb.Airplane(
            fuselages=[fuselage],
            xyz_ref=xyz_ref
        ),
        op_point=asb.OperatingPoint(
            velocity=10,
            alpha=5,
            beta=5
        )
    ).run()


x_cgs = np.linspace(0, 1, 11)
aeros = np.array([
    get_aero(xyz_ref=[x, 0, 0])
    for x in x_cgs
], dtype="O")

import matplotlib.pyplot as plt
import aerosandbox.tools.pretty_plots as p

fig, ax = plt.subplots(figsize=(4, 4))

plt.plot(
    x_cgs,
    np.array([a['m_b'] for a in aeros]),
    ".-"
)
p.show_plot(
    "Fuselage Pitching Moment",
    r"$x_{cg} / l$",
    "Pitching Moment [Nm]"
)
"""
Expected result:

For CG far forward (0), vehicle should have negative pitching moment.

For CG far aft (1), vehicle should have positive pitching moment.
"""



from aerosandbox.dynamics.rigid_body.rigid_3D.body_euler import DynamicsRigidBody3DBodyEuler
from aerosandbox.weights.mass_properties import MassProperties
import aerosandbox.numpy as np
from typing import Union, Dict, Tuple


class DynamicsRigidBody2DBody(DynamicsRigidBody3DBodyEuler):
    """
    Dynamics instance:
    * simulating a rigid body
    * in 2D
    * with velocity parameterized in body axes

    State variables:
        x_e: x-position, in Earth axes. [meters]
        z_e: z-position, in Earth axes. [meters]
        u_b: x-velocity, in body axes. [m/s]
        w_b: z-velocity, in body axes. [m/s]
        theta: pitch angle. [rad]
        q: y-angular-velocity, in body axes. [rad/sec]

    Control variables:
        Fx_b: Force along the body-x axis. [N]
        Fz_b: Force along the body-z axis. [N]
        My_b: Moment about the body-y axis. [Nm]

    """

    def __init__(self,
                 mass_props: MassProperties = None,
                 x_e: Union[float, np.ndarray] = 0,
                 z_e: Union[float, np.ndarray] = 0,
                 u_b: Union[float, np.ndarray] = 0,
                 w_b: Union[float, np.ndarray] = 0,
                 theta: Union[float, np.ndarray] = 0,
                 q: Union[float, np.ndarray] = 0,
                 ):
        # Initialize state variables
        self.mass_props = MassProperties() if mass_props is None else mass_props
        self.x_e = x_e
        self.y_e = 0
        self.z_e = z_e
        self.u_b = u_b
        self.v_b = 0
        self.w_b = w_b
        self.phi = 0
        self.theta = theta
        self.psi = 0
        self.p = 0
        self.q = q
        self.r = 0

        # Initialize control variables
        self.Fx_b = 0
        self.Fy_b = 0
        self.Fz_b = 0
        self.Mx_b = 0
        self.My_b = 0
        self.Mz_b = 0
        self.hx_b = 0
        self.hy_b = 0
        self.hz_b = 0

    @property
    def state(self):
        return {
            "x_e"  : self.x_e,
            "z_e"  : self.z_e,
            "u_b"  : self.u_b,
            "w_b"  : self.w_b,
            "theta": self.theta,
            "q"    : self.q,
        }

    @property
    def control_variables(self):
        return {
            "Fx_b": self.Fx_b,
            "Fz_b": self.Fz_b,
            "My_b": self.My_b,
        }

    def state_derivatives(self) -> Dict[str, Union[float, np.ndarray]]:
        derivatives = super().state_derivatives()
        return {
            k: derivatives[k] for k in self.state.keys()
        }


if __name__ == '__main__':
    dyn = DynamicsRigidBody2DBody()



import warnings
from math import atan2, degrees

import matplotlib.patheffects as path_effects
import numpy as np
from matplotlib.container import ErrorbarContainer
from matplotlib.dates import DateConverter, num2date

from .utils import ensure_float, maximum_bipartite_matching, always_iterable


# Label line with line2D label data
def labelLine(
        line,
        x,
        label=None,
        align=True,
        drop_label=False,
        yoffset=0,
        yoffset_logspace=False,
        outline_color="auto",
        outline_width=8,
        **kwargs,
):
    """Label a single matplotlib line at position x

    Parameters
    ----------
    line : matplotlib.lines.Line
       The line holding the label
    x : number
       The location in data unit of the label
    label : string, optional
       The label to set. This is inferred from the line by default
    drop_label : bool, optional
       If True, the label is consumed by the function so that subsequent
       calls to e.g. legend do not use it anymore.
    yoffset : double, optional
        Space to add to label's y position
    yoffset_logspace : bool, optional
        If True, then yoffset will be added to the label's y position in
        log10 space
    outline_color : None | "auto" | color
        Colour of the outline. If set to "auto", use the background color.
        If set to None, do not draw an outline.
    outline_width : number
        Width of the outline
    kwargs : dict, optional
       Optional arguments passed to ax.text
    """

    ax = line.axes
    xdata = ensure_float(line.get_xdata())
    ydata = line.get_ydata()

    mask = np.isfinite(ydata)
    if mask.sum() == 0:
        raise Exception(f"The line {line} only contains nan!")

    # Find first segment of xdata containing x
    if isinstance(xdata, tuple) and len(xdata) == 2:
        i = 0
        xa = min(xdata)
        xb = max(xdata)
    else:
        for imatch, (xa, xb) in enumerate(zip(xdata[:-1], xdata[1:])):
            if min(xa, xb) <= ensure_float(x) <= max(xa, xb):
                i = imatch
                break
        else:
            raise Exception("x label location is outside data range!")

    xfa = ensure_float(xa)
    xfb = ensure_float(xb)
    ya = ydata[i]
    yb = ydata[i + 1]

    # Handle vertical case
    if xfb == xfa:
        fraction = 0.5
    else:
        fraction = (ensure_float(x) - xfa) / (xfb - xfa)

    if yoffset_logspace:
        y = ya + (yb - ya) * fraction
        y *= 10 ** yoffset
    else:
        y = ya + (yb - ya) * fraction + yoffset

    if not (np.isfinite(ya) and np.isfinite(yb)):
        warnings.warn(
            (
                "%s could not be annotated due to `nans` values. "
                "Consider using another location via the `x` argument."
            )
            % line,
            UserWarning,
        )
        return

    if not label:
        label = line.get_label()

    if drop_label:
        line.set_label(None)

    if align:
        # Compute the slope and label rotation
        screen_dx, screen_dy = ax.transData.transform(
            (xfa, ya)
        ) - ax.transData.transform((xfb, yb))
        rotation = (degrees(atan2(screen_dy, screen_dx)) + 90) % 180 - 90
    else:
        rotation = 0

    # Set a bunch of keyword arguments
    if "color" not in kwargs:
        kwargs["color"] = line.get_color()

    if ("horizontalalignment" not in kwargs) and ("ha" not in kwargs):
        kwargs["ha"] = "center"

    if ("verticalalignment" not in kwargs) and ("va" not in kwargs):
        kwargs["va"] = "center"

    if "clip_on" not in kwargs:
        kwargs["clip_on"] = True

    if "zorder" not in kwargs:
        kwargs["zorder"] = 2.5

    if outline_color == "auto":
        outline_color = ax.get_facecolor()

    txt = ax.text(x, y, label, rotation=rotation, **kwargs)

    if outline_color is None:
        effects = [path_effects.Normal()]
    else:
        effects = [
            path_effects.Stroke(linewidth=outline_width, foreground=outline_color),
            path_effects.Normal(),
        ]

    txt.set_path_effects(effects)
    return txt


def labelLines(
        lines,
        align=True,
        xvals=None,
        drop_label=False,
        shrink_factor=0.05,
        yoffsets=0,
        outline_color="auto",
        outline_width=5,
        **kwargs,
):
    """Label all lines with their respective legends.

    Parameters
    ----------
    lines : list of matplotlib lines
       The lines to label
    align : boolean, optional
       If True, the label will be aligned with the slope of the line
       at the location of the label. If False, they will be horizontal.
    xvals : (xfirst, xlast) or array of float, optional
       The location of the labels. If a tuple, the labels will be
       evenly spaced between xfirst and xlast (in the axis units).
    drop_label : bool, optional
       If True, the label is consumed by the function so that subsequent
       calls to e.g. legend do not use it anymore.
    shrink_factor : double, optional
       Relative distance from the edges to place closest labels. Defaults to 0.05.
    yoffsets : number or list, optional.
        Distance relative to the line when positioning the labels. If given a number,
        the same value is used for all lines.
    outline_color : None | "auto" | color
        Colour of the outline. If set to "auto", use the background color.
        If set to None, do not draw an outline.
    outline_width : number
        Width of the outline
    kwargs : dict, optional
       Optional arguments passed to ax.text
    """
    ax = lines[0].axes

    handles, allLabels = ax.get_legend_handles_labels()

    all_lines = []
    for h in handles:
        if isinstance(h, ErrorbarContainer):
            all_lines.append(h.lines[0])
        else:
            all_lines.append(h)

    # In case no x location was provided, we need to use some heuristics
    # to generate them.
    if xvals is None:
        xvals = ax.get_xlim()
        xvals_rng = xvals[1] - xvals[0]
        shrinkage = xvals_rng * shrink_factor
        xvals = (xvals[0] + shrinkage, xvals[1] - shrinkage)

    if isinstance(xvals, tuple) and len(xvals) == 2:
        xmin, xmax = xvals
        xscale = ax.get_xscale()
        if xscale == "log":
            xvals = np.logspace(np.log10(xmin), np.log10(xmax), len(all_lines) + 2)[
                    1:-1
                    ]
        else:
            xvals = np.linspace(xmin, xmax, len(all_lines) + 2)[1:-1]

        # Build matrix line -> xvalue
        ok_matrix = np.zeros((len(all_lines), len(all_lines)), dtype=bool)

        for i, line in enumerate(all_lines):
            xdata = ensure_float(line.get_xdata())
            minx, maxx = min(xdata), max(xdata)
            for j, xv in enumerate(xvals):
                ok_matrix[i, j] = minx < xv < maxx

        # If some xvals do not fall in their corresponding line,
        # find a better matching using maximum bipartite matching.
        if not np.all(np.diag(ok_matrix)):
            order = maximum_bipartite_matching(ok_matrix)

            # The maximum match may miss a few points, let's add them back
            imax = order.max()
            order[order < 0] = np.arange(imax + 1, len(order))

            # Now reorder the xvalues
            old_xvals = xvals.copy()
            xvals[order] = old_xvals
    else:
        xvals = list(always_iterable(xvals))  # force the creation of a copy

    labLines, labels = [], []
    # Take only the lines which have labels other than the default ones
    for i, (line, xv) in enumerate(zip(all_lines, xvals)):
        label = allLabels[all_lines.index(line)]
        labLines.append(line)
        labels.append(label)

        # Move xlabel if it is outside valid range
        xdata = ensure_float(line.get_xdata())
        if not (min(xdata) <= xv <= max(xdata)):
            warnings.warn(
                (
                    "The value at position %s in `xvals` is outside the range of its "
                    "associated line (xmin=%s, xmax=%s, xval=%s). Clipping it "
                    "into the allowed range."
                )
                % (i, min(xdata), max(xdata), xv),
                UserWarning,
            )
            new_xv = min(xdata) + (max(xdata) - min(xdata)) * 0.9
            xvals[i] = new_xv

    # Convert float values back to datetime in case of datetime axis
    if isinstance(ax.xaxis.converter, DateConverter):
        xvals = [num2date(x).replace(tzinfo=ax.xaxis.get_units()) for x in xvals]

    txts = []
    try:
        yoffsets = [float(yoffsets)] * len(all_lines)
    except TypeError:
        pass
    for line, x, yoffset, label in zip(labLines, xvals, yoffsets, labels):
        txts.append(
            labelLine(
                line,
                x,
                label=label,
                align=align,
                drop_label=drop_label,
                yoffset=yoffset,
                outline_color=outline_color,
                outline_width=outline_width,
                **kwargs,
            )
        )

    return txts



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest
import matplotlib.pyplot as plt

"""
Hanging Chain problem from https://web.casadi.org/blog/opti/

Next, we will visit the hanging chain problem. We consider 
N point masses, connected by springs, hung from two fixed points at (-2, 1) and (2, 1), subject to gravity.

The chain should also remain above a locally-concave ground surface.

We seek the rest position of the system, obtained by minimizing the total energy in the system.

"""


def test_opti_hanging_chain_with_callback(plot=False):
    N = 40
    m = 40 / N
    D = 70 * N
    g = 9.81
    L = 1

    opti = asb.Opti()

    x = opti.variable(
        init_guess=np.linspace(-2, 2, N)
    )
    y = opti.variable(
        init_guess=1,
        n_vars=N,
    )

    distance = np.sqrt(  # Distance from one node to the next
        np.diff(x) ** 2 + np.diff(y) ** 2
    )

    potential_energy_spring = 0.5 * D * np.sum((distance - L / N) ** 2)
    potential_energy_gravity = g * m * np.sum(y)
    potential_energy = potential_energy_spring + potential_energy_gravity

    opti.minimize(potential_energy)

    # Add end point constraints
    opti.subject_to([
        x[0] == -2,
        y[0] == 1,
        x[-1] == 2,
        y[-1] == 1
    ])

    # Add a ground constraint
    opti.subject_to(
        y >= np.cos(0.1 * x) - 0.5
    )

    # Add a callback

    if plot:
        def my_callback(iter: int):
            plt.plot(
                opti.debug.value(x),
                opti.debug.value(y),
                ".-",
                label=f"Iter {iter}",
                zorder=3 + iter
            )

        fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
        x_ground = np.linspace(-2, 2, N)
        y_ground = np.cos(0.1 * x_ground) - 0.5
        plt.plot(x_ground, y_ground, "--k", zorder=2)

    else:
        def my_callback(iter: int):
            print(f"Iter {iter}")
            print(f"\tx = {opti.debug.value(x)}")
            print(f"\ty = {opti.debug.value(y)}")

    sol = opti.solve(
        callback=my_callback
    )

    assert sol.value(potential_energy) == pytest.approx(626.462, abs=1e-3)

    if plot:
        plt.show()


if __name__ == '__main__':
    pytest.main()



import aerosandbox.numpy as np
from aerosandbox.optimization.opti import Opti
from abc import abstractmethod
import copy
from typing import Dict, Any
import casadi as cas


class AeroSandboxObject:

    @abstractmethod
    def __init__(self):
        """
        Denotes AeroSandboxObject as an abstract class, meaning you can't instantiate it directly - you must subclass
        (extend) it instead.
        """
        pass

    @abstractmethod
    def __repr__(self):
        """
        Forces all subclasses to implement a __repr__ method, which is good practice.
        """

    def copy(self):
        """
        Returns a shallow copy of the object.
        """
        return copy.copy(self)

    def deepcopy(self):
        """
        Returns a deep copy of the object.
        """
        return copy.deepcopy(self)

    def substitute_solution(self,
                            sol: cas.OptiSol,
                            inplace: bool = None,
                            ):
        """
        Substitutes a solution from CasADi's solver recursively as an in-place operation.

        In-place operation. To make it not in-place, do `y = copy.deepcopy(x)` or similar first.
        :param sol: OptiSol object.
        :return:
        """
        import warnings
        warnings.warn(
            "This function is deprecated and will break at some future point.\n"
            "Use `sol.value(x)` or even simply `sol(x)` instead, which now works\n"
            "recursively on complex data structures.",
            DeprecationWarning
        )

        # Set defaults
        if inplace is None:
            inplace = True

        def convert(item):
            """
            This is essentially a supercharged version of sol.value(), which works for more iterable types.

            Args:
                item:

            Returns:

            """

            # If it can be converted, do the conversion.
            if np.is_casadi_type(item, recursive=False):
                return sol.value(item)

            t = type(item)

            # If it's a Python iterable, recursively convert it, and preserve the type as best as possible.
            if issubclass(t, list):
                return [convert(i) for i in item]
            if issubclass(t, tuple):
                return tuple([convert(i) for i in item])
            if issubclass(t, set) or issubclass(t, frozenset):
                return {convert(i) for i in item}
            if issubclass(t, dict):
                return {
                    convert(k): convert(v)
                    for k, v in item.items()
                }

            # Skip certain Python types
            for type_to_skip in (
                    bool, str,
                    int, float, complex,
                    range,
                    type(None),
                    bytes, bytearray, memoryview
            ):
                if issubclass(t, type_to_skip):
                    return item

            # Skip certain CasADi types
            for type_to_skip in (
                    cas.Opti,
                    cas.OptiSol
            ):
                if issubclass(t, type_to_skip):
                    return item

            # If it's any other type, try converting its attribute dictionary:
            try:
                newdict = {
                    k: convert(v)
                    for k, v in item.__dict__.items()
                }

                if inplace:
                    for k, v in newdict.items():
                        setattr(item, k, v)

                    return item

                else:
                    newitem = copy.copy(item)
                    for k, v in newdict.items():
                        setattr(newitem, k, v)

                    return newitem

            except AttributeError:
                pass

            # Try converting it blindly. This will catch most NumPy-array-like types.
            try:
                return sol.value(item)
            except (NotImplementedError, TypeError, ValueError):
                pass

            # At this point, we're not really sure what type the object is. Raise a warning and return the item, then hope for the best.
            import warnings
            warnings.warn(f"In solution substitution, could not convert an object of type {t}.\n"
                          f"Returning it and hoping for the best.", UserWarning)

            return item

        if inplace:
            convert(self)

        else:
            return convert(self)


class ExplicitAnalysis(AeroSandboxObject):
    default_analysis_specific_options: Dict[type, Dict[str, Any]] = {}
    """This is part of AeroSandbox's "analysis-specific options" feature, which lets you "tag" geometry objects with 
    flags that change how different analyses act on them. 
    
    This variable, `default_analysis_specific_options`, allows you to specify default values for options that can be used for 
    specific problems. 
    
    This should be a dictionary, where: * keys are the geometry-like types that you might be interested in defining 
    parameters for. * values are dictionaries, where: * keys are strings that label a given option * values are 
    anything. These are used as the default values, in the event that the associated geometry doesn't override those. 
    
    An example of what this variable might look like, for a vortex-lattice method aerodynamic analysis:
    
    >>> default_analysis_specific_options = {
    >>>     Airplane: dict(
    >>>         profile_drag_coefficient=0
    >>>     ),
    >>>     Wing    : dict(
    >>>         wing_level_spanwise_spacing=True,
    >>>         spanwise_resolution=12,
    >>>         spanwise_spacing="cosine",
    >>>         chordwise_resolution=12,
    >>>         chordwise_spacing="cosine",
    >>>         component=None,  # type: int
    >>>         no_wake=False,
    >>>         no_alpha_beta=False,
    >>>         no_load=False,
    >>>         drag_polar=dict(
    >>>             CL1=0,
    >>>             CD1=0,
    >>>             CL2=0,
    >>>             CD2=0,
    >>>             CL3=0,
    >>>             CD3=0,
    >>>         ),
    >>>     )
    >>> }
    
    """

    def get_options(self,
                    geometry_object: AeroSandboxObject,
                    ) -> Dict[str, Any]:
        """
        Retrieves the analysis-specific options that correspond to both:

            * An analysis type (which is this object, "self"), and

            * A specific geometry object, such as an Airplane or Wing.

        Args:
            geometry_object: An instance of an AeroSandbox geometry object, such as an Airplane, Wing, etc.

                * In order for this function to do something useful, you probably want this option to have
                `analysis_specific_options` defined. See the asb.Airplane constructor for an example of this.

        Returns: A dictionary that combines:

            * This analysis's default options for this geometry, if any exist.

            * The geometry's declared analysis-specific-options for this analysis, if it exists. These geometry
            options will override the defaults from the analysis.

            This dictionary has the format:

                * keys are strings, listing specific options

                * values can be any type, and simply state the value of the analysis-specific option following the
                logic above.

        Note: if this analysis defines a set of default options for the geometry type in question (by using
        `self.default_analysis_specific_options`), all keys from the geometry object's `analysis_specific_options`
        will be validated against those in the default options set. A warning will be raised if keys do not
        correspond to those in the defaults, as this (potentially) indicates a typo, which would otherwise be
        difficult to debug.

        """

        ### Determine the types of both this analysis and the geometry object.
        analysis_type: type = self.__class__
        geometry_type: type = geometry_object.__class__

        ### Determine whether this analysis and the geometry object have options that specifically reference each other or not.
        try:
            analysis_options_for_this_geometry = self.default_analysis_specific_options[geometry_type]
            assert hasattr(analysis_options_for_this_geometry, "items")
        except (AttributeError, KeyError, AssertionError):
            analysis_options_for_this_geometry = None

        try:
            geometry_options_for_this_analysis = geometry_object.analysis_specific_options[analysis_type]
            assert hasattr(geometry_options_for_this_analysis, "items")
        except (AttributeError, KeyError, AssertionError):
            geometry_options_for_this_analysis = None

        ### Now, merge those options (with logic depending on whether they exist or not)
        if analysis_options_for_this_geometry is not None:
            options = copy.deepcopy(analysis_options_for_this_geometry)
            if geometry_options_for_this_analysis is not None:
                for k, v in geometry_options_for_this_analysis.items():
                    if k in analysis_options_for_this_geometry.keys():
                        options[k] = v
                    else:

                        import warnings
                        allowable_keys = [f'"{k}"' for k in analysis_options_for_this_geometry.keys()]
                        warnings.warn(
                            f"\nAn object of type '{geometry_type.__name__}' declared the analysis-specific option '{k}' for use with analysis '{analysis_type.__name__}'.\n"
                            f"This was unexpected! Allowable analysis-specific options for '{geometry_type.__name__}' with '{analysis_type.__name__}' are:\n"
                            "\t" + "\n\t".join(allowable_keys) + "\n" "Did you make a typo?",
                            stacklevel=2,
                        )

        else:
            if geometry_options_for_this_analysis is not None:
                options = geometry_options_for_this_analysis
            else:
                options = {}

        return options


class ImplicitAnalysis(AeroSandboxObject):

    @staticmethod
    def initialize(init_method):
        """
        A decorator that should be applied to the __init__ method of ImplicitAnalysis or any subclass of it.

        Usage example:

        >>> class MyAnalysis(ImplicitAnalysis):
        >>>
        >>>     @ImplicitAnalysis.initialize
        >>>     def __init__(self):
        >>>         self.a = self.opti.variable(init_guess = 1)
        >>>         self.b = self.opti.variable(init_guess = 2)
        >>>
        >>>         self.opti.subject_to(
        >>>             self.a == self.b ** 2
        >>>         ) # Add a nonlinear governing equation

        Functionality:

        The basic purpose of this wrapper is to ensure that every ImplicitAnalysis has an `opti` property that points to
        an optimization environment (asb.Opti type) that it can work in.

        How do we obtain an asb.Opti environment to work in? Well, this decorator adds an optional `opti` parameter to
        the __init__ method that it is applied to.

            1. If this `opti` parameter is not provided, then a new empty `asb.Opti` environment is created and stored as
            `ImplicitAnalysis.opti`.

            2. If the `opti` parameter is provided, then we simply assign the given `asb.Opti` environment (which may
            already contain other variables/constraints/objective) to `ImplicitAnalysis.opti`.

        In addition, a property called `ImplicitAnalysis.opti_provided` is stored, which records whether the user
        provided an Opti environment or if one was instead created for them.

        If the user did not provide an Opti environment (Option 1 from our list above), we assume that the user basically
        just wants to perform a normal, single-disciplinary analysis. So, in this case, we proceed to solve the analysis as-is
        and do an in-place substitution of the solution.

        If the user did provide an Opti environment (Option 2 from our list above), we assume that the user might potentially want
        to add other implicit analyses to the problem. So, in this case, we don't solve the analysis, and the user must later
        solve the analysis by calling `sol = opti.solve()` or similar.

        """

        def init_wrapped(self, *args, opti=None, **kwargs):
            if opti is None:
                self.opti = Opti()
                self.opti_provided = False
            else:
                self.opti = opti
                self.opti_provided = True

            init_method(self, *args, **kwargs)

            if not self.opti_provided and not self.opti.x.shape == (0, 1):
                sol = self.opti.solve()
                self.substitute_solution(sol)

        return init_wrapped

    class ImplicitAnalysisInitError(Exception):
        def __init__(self,
                     message="""
    Your ImplicitAnalysis object doesn't have an `opti` property!
    This is almost certainly because you didn't decorate your object's __init__ method with 
    `@ImplicitAnalysis.initialize`, which you should go do.
                     """
                     ):
            self.message = message
            super().__init__(self.message)

    @property
    @abstractmethod
    def opti(self):
        try:
            return self._opti
        except AttributeError:
            raise self.ImplicitAnalysisInitError()

    @opti.setter
    @abstractmethod
    def opti(self, value: Opti):
        self._opti = value

    @property
    @abstractmethod
    def opti_provided(self):
        try:
            return self._opti_provided
        except AttributeError:
            raise self.ImplicitAnalysisInitError()

    @opti_provided.setter
    @abstractmethod
    def opti_provided(self, value: bool):
        self._opti_provided = value



import aerosandbox.numpy as np
import pandas as pd
from pathlib import Path

### Define constants
gas_constant_universal = 8.31432  # J/(mol*K); universal gas constant
molecular_mass_air = 28.9644e-3  # kg/mol; molecular mass of air
gas_constant_air = gas_constant_universal / molecular_mass_air  # J/(kg*K); gas constant of air
g = 9.81  # m/s^2, gravitational acceleration on earth

### Read ISA table data
isa_table = pd.read_csv(Path(__file__).parent.absolute() / "isa_data/isa_table.csv")
isa_base_altitude = isa_table["Base Altitude [m]"].values
isa_lapse_rate = isa_table["Lapse Rate [K/km]"].values / 1000
isa_base_temperature = isa_table["Base Temperature [C]"].values + 273.15


### Calculate pressure at each ISA level programmatically using the barometric pressure equation with linear temperature.
def barometric_formula(
        P_b,
        T_b,
        L_b,
        h,
        h_b,
):
    """
    The barometric pressure equation, from here: https://en.wikipedia.org/wiki/Barometric_formula
    Args:
        P_b: Pressure at the base of the layer, in Pa
        T_b: Temperature at the base of the layer, in K
        L_b: Temperature lapse rate, in K/m
        h: Altitude, in m
        h_b:

    Returns:

    """
    T = T_b + L_b * (h - h_b)
    T = np.fmax(T, 1)  # Keep temperature nonnegative, no matter the inputs.
    if L_b != 0:
        return P_b * (T / T_b) ** (-g / (gas_constant_air * L_b))
    else:
        return P_b * np.exp(
            np.clip(
                -g * (h - h_b) / (gas_constant_air * T_b),
                -500,
                500
            )
        )


isa_pressure = [101325.]  # Pascals
for i in range(len(isa_table) - 1):
    isa_pressure.append(
        barometric_formula(
            P_b=isa_pressure[i],
            T_b=isa_base_temperature[i],
            L_b=isa_lapse_rate[i],
            h=isa_base_altitude[i + 1],
            h_b=isa_base_altitude[i]
        )
    )


def pressure_isa(altitude):
    """
    Computes the pressure at a given altitude based on the International Standard Atmosphere.

    Uses the Barometric formula, as implemented here: https://en.wikipedia.org/wiki/Barometric_formula

    Args:
        altitude: Geopotential altitude [m]

    Returns: Pressure [Pa]

    """
    pressure = 0 * altitude  # Initialize the pressure to all zeros.

    for i in range(len(isa_table)):
        pressure = np.where(
            altitude > isa_base_altitude[i],
            barometric_formula(
                P_b=isa_pressure[i],
                T_b=isa_base_temperature[i],
                L_b=isa_lapse_rate[i],
                h=altitude,
                h_b=isa_base_altitude[i]
            ),
            pressure
        )

    ### Add lower bound case
    pressure = np.where(
        altitude <= isa_base_altitude[0],
        barometric_formula(
            P_b=isa_pressure[0],
            T_b=isa_base_temperature[0],
            L_b=isa_lapse_rate[0],
            h=altitude,
            h_b=isa_base_altitude[0]
        ),
        pressure
    )

    return pressure


def temperature_isa(altitude):
    """
    Computes the temperature at a given altitude based on the International Standard Atmosphere.

    Args:
        altitude: Geopotential altitude [m]

    Returns: Temperature [K]

    """
    temp = 0 * altitude  # Initialize the temperature to all zeros.

    for i in range(len(isa_table)):
        temp = np.where(
            altitude > isa_base_altitude[i],
            (altitude - isa_base_altitude[i]) * isa_lapse_rate[i] + isa_base_temperature[i],
            temp
        )

    ### Add lower bound case
    temp = np.where(
        altitude <= isa_base_altitude[0],
        (altitude - isa_base_altitude[0]) * isa_lapse_rate[0] + isa_base_temperature[0],
        temp
    )

    return temp


if __name__ == '__main__':
    pressure_isa(-50e3)



from aerosandbox.common import AeroSandboxObject
import aerosandbox.numpy as np
from aerosandbox.atmosphere._isa_atmo_functions import pressure_isa, temperature_isa
from aerosandbox.atmosphere._diff_atmo_functions import pressure_differentiable, temperature_differentiable
import aerosandbox.tools.units as u

### Define constants
gas_constant_universal = 8.31432  # J/(mol*K); universal gas constant
molecular_mass_air = 28.9644e-3  # kg/mol; molecular mass of air
gas_constant_air = gas_constant_universal / molecular_mass_air  # J/(kg*K); gas constant of air
effective_collision_diameter = 0.365e-9  # m, effective collision diameter of an air molecule


### Define the Atmosphere class
class Atmosphere(AeroSandboxObject):
    r"""
    All models here are smoothed fits to the 1976 COESA model;
    see AeroSandbox\studies\Atmosphere Fitting for details.

    """

    def __init__(self,
                 altitude: float = 0.,  # meters
                 method: str = "differentiable"
                 ):
        """
        Initialize a new Atmosphere.
        
        Args:
            
            altitude: Flight altitude, in meters. This is assumed to be a geopotential altitude above MSL.
            
            method: Method of atmosphere modeling to use. Either:
                * "differentiable" - a C1-continuous fit to the International Standard Atmosphere
                * "isa" - the International Standard Atmosphere
                
        """
        self.altitude = altitude
        self.method = method
        self._valid_altitude_range = (0, 80000)

    def __repr__(self) -> str:
        try:
            altitude_string = f"altitude: {self.altitude:.0f} m ({self.altitude / u.foot:.0f} ft)"
        except (ValueError, TypeError):
            altitude_string = f"altitude: {self.altitude} m"

        return f"Atmosphere ({altitude_string}, method: '{self.method}')"

    ### The two primary state variables, pressure and temperature, go here!

    def pressure(self):
        """
        Returns the pressure, in Pascals.
        """
        if self.method.lower() == "isa":
            return pressure_isa(self.altitude)
        elif self.method.lower() == "differentiable":
            return pressure_differentiable(self.altitude)
        else:
            raise ValueError("Bad value of 'type'!")

    def temperature(self):
        """
        Returns the temperature, in Kelvin.
        """
        if self.method.lower() == "isa":
            return temperature_isa(self.altitude)
        elif self.method.lower() == "differentiable":
            return temperature_differentiable(self.altitude)
        else:
            raise ValueError("Bad value of 'type'!")

    ### Everything else in this class is a derived quantity; all models of derived quantities go here.

    def density(self):
        """
        Returns the density, in kg/m^3.
        """
        rho = self.pressure() / (self.temperature() * gas_constant_air)

        return rho

    def speed_of_sound(self):
        """
        Returns the speed of sound, in m/s.
        """
        temperature = self.temperature()
        return (self.ratio_of_specific_heats() * gas_constant_air * temperature) ** 0.5

    def dynamic_viscosity(self):
        """
        Returns the dynamic viscosity (mu), in kg/(m*s).

        Based on Sutherland's Law, citing `https://www.cfd-online.com/Wiki/Sutherland's_law`.

        According to Rathakrishnan, E. (2013). Theoretical aerodynamics. John Wiley & Sons.:
        This relationship is valid from 0.01 to 100 atm, and between 0 and 3000K.

        According to White, F. M., & Corfield, I. (2006). Viscous fluid flow (Vol. 3, pp. 433-434). New York: McGraw-Hill.:
        The error is no more than approximately 2% for air between 170K and 1900K.
        """

        # Sutherland constants
        C1 = 1.458e-6  # kg/(m*s*sqrt(K))
        S = 110.4  # K

        # Sutherland equation
        temperature = self.temperature()
        mu = C1 * temperature ** 1.5 / (temperature + S)

        return mu

    def kinematic_viscosity(self):
        """
        Returns the kinematic viscosity (nu), in m^2/s.

        Definitional.
        """
        return self.dynamic_viscosity() / self.density()

    def ratio_of_specific_heats(self):
        return 1.4  # TODO model temperature variation

    # def thermal_velocity(self):
    #     """
    #     Returns the thermal velocity (mean particle speed)
    #     Returns:
    #
    #     """
    #
    def mean_free_path(self):
        """
        Returns the mean free path of an air molecule, in meters.

        To find the collision radius, assumes "a hard-sphere gas that has the same viscosity as the actual gas being considered".

        From Vincenti, W. G. and Kruger, C. H. (1965). Introduction to physical gas dynamics. Krieger Publishing Company. p. 414.

        """
        return self.dynamic_viscosity() / self.pressure() * np.sqrt(
            np.pi * gas_constant_universal * self.temperature() / (2 * molecular_mass_air)
        )

    def knudsen(self, length):
        """
        Computes the Knudsen number for a given length.
        """
        return self.mean_free_path() / length


if __name__ == "__main__":
    # Make AeroSandbox Atmosphere
    altitude = np.linspace(-5e3, 100e3, 1000)
    atmo_diff = Atmosphere(altitude=altitude)
    atmo_isa = Atmosphere(altitude=altitude, method="isa")

    from aerosandbox.tools.pretty_plots import plt, sns, mpl, show_plot, set_ticks

    fig, ax = plt.subplots()

    plt.plot(
        (
                (atmo_diff.pressure() - atmo_isa.pressure()) / atmo_isa.pressure()
        ) * 100,
        altitude / 1e3,
    )
    set_ticks(0.2, 0.1, 20, 10)
    plt.xlim(-1, 1)
    show_plot(
        "AeroSandbox Atmosphere vs. ISA Atmosphere",
        "Pressure, Relative Error [%]",
        "Altitude [km]"
    )

    fig, ax = plt.subplots()
    plt.plot(
        atmo_diff.temperature() - atmo_isa.temperature(),
        altitude / 1e3,
    )
    set_ticks(1, 0.5, 20, 10)
    plt.xlim(-5, 5)
    show_plot(
        "AeroSandbox Atmosphere vs. ISA Atmosphere",
        "Temperature, Absolute Error [K]",
        "Altitude [km]"
    )



import aerosandbox.numpy as np


def column_buckling_critical_load(
        elastic_modulus: float,
        moment_of_inertia: float,
        length: float,
        boundary_condition_type: str = "pin-pin",
        use_recommended_design_values: bool = True,
):
    """
    Computes the critical load (in N) for a column or tube in compression to buckle via primary buckling. Uses Euler's classical critical
    load formula.

    Args:
        elastic_modulus: The elastic modulus of the material, in Pa.

        moment_of_inertia: The moment of inertia of the cross-section, in m^4.

        length: The length of the column, in m.

        boundary_condition_type: The boundary condition type. Options are:
            - "pin-pin"
            - "pin-clamp"
            - "clamp-clamp"
            - "clamp-pin"
            - "clamp-free"
            - "free-clamp"

        use_recommended_design_values: Whether to use the recommended design value of K for a given boundary condition (True)
        or to use the less-conservative theoretical value (False).

            * Recommended values are from Table C.1.8.1 in Steel Construction Manual, 8th edition, 2nd revised
            printing, American Institute of Steel Construction, 1987 via WikiMedia:
            https://commons.wikimedia.org/wiki/File:ColumnEffectiveLength.png

    Returns:
        The critical compressive load (in N) for the column or tube to buckle via primary buckling.

    """
    if boundary_condition_type == "pin-pin":
        K = 1.00 if use_recommended_design_values else 1.00
    elif boundary_condition_type == "pin-clamp" or boundary_condition_type == "clamp-pin":
        K = 0.80 if use_recommended_design_values else 0.70
    elif boundary_condition_type == "clamp-clamp":
        K = 0.65 if use_recommended_design_values else 0.50
    elif boundary_condition_type == "clamp-free" or boundary_condition_type == "free-clamp":
        K = 2.10 if use_recommended_design_values else 2.00
    else:
        raise ValueError("Invalid `boundary_condition_type`.")

    return (
            np.pi ** 2 * elastic_modulus * moment_of_inertia
            / (K * length) ** 2
    )


def thin_walled_tube_crippling_buckling_critical_load(
        elastic_modulus: float,
        wall_thickness: float,
        radius: float,
        use_recommended_design_values: bool = True,
):
    """
    Computes the critical load for a thin-walled tube in compression to fail in the crippling mode. (Note: you should also check for
    failure by primary buckling using the `column_buckling_critical_load()` function.)

    The crippling mode is a specific instability mode for tubes with thin walls when loaded in compression. It can be
    seen when you step on a soda can and it buckles inwards. The critical load for this mode is given by the
    following formula:

        stress_crippling = crippling_constant * (E * t / r)

    where:

        A recommended value of crippling_constant = 0.3 is given in Raymer: Aircraft Design: A Conceptual Approach,
        5th Edition, Eq. 14.33, pg. 554.

        A theoretically more accurate value of crippling_constant = 0.605 is given in the Air Force Stress Manual,
        Section 2.3.2.1, Eq. 2-20. This value assumes mu = 0.3, which is a good assumption for most metals.

        and E is the elastic modulus, t is the wall thickness, and r is the radius.

    For more info, see the Air Force Stress Manual, Section 2.3.2.1:
    https://engineeringlibrary.org/reference/column-crippling-air-force-stress-manual

    And see Raymer: Aircraft Design: A Conceptual Approach, 5th Edition, pg. 554.

    Args:
        elastic_modulus: The elastic modulus of the material, in Pa.

        wall_thickness: The wall thickness of the tube, in m.

        radius: The radius of the tube, in m.

        use_recommended_design_values: Whether to use the recommended design value of crippling_constant (True)
        or to use the less-conservative theoretical value (False).

    Returns:
        The critical compressive load (in N) for the tube to buckle in the crippling mode.

    """
    if use_recommended_design_values:
        crippling_stress_constant = 0.3
        # Taken from Raymer: Aircraft Design: A Conceptual Approach, 5th Edition, Eq. 14.33, pg. 554.
        #
        # According to the Air Force Stress Manual, Figure 2-67, this value should drop as radius/wall_thickness
        # increases.
    else:
        crippling_stress_constant = 0.605
        # Theoretically, this should be (3 * (1 - mu^2))^(-0.5), where mu is the Poisson's ratio.
        # Following the Air Force Stress Manual, Section 2.3.2.1, Eq. 2-20.
        # The above value assumes mu = 0.3, which is a good assumption for most metals.

    crippling_stress = 0.3 * (elastic_modulus * wall_thickness / radius)

    tube_xsec_area = 2 * np.pi * radius * wall_thickness

    crippling_load = crippling_stress * tube_xsec_area

    return crippling_load


def plate_buckling_critical_load(
        length: float,
        width: float,
        wall_thickness: float,
        elastic_modulus: float,
        side_boundary_condition_type: str = "clamp-clamp",
):
    """
    Computes the critical compressive load (in N) for a plate to buckle via plate buckling.

    Assumes a rectangular plate with dimensions:
    - length
    - width
    - wall_thickness

    A compressive force is applied such that it is aligned with the length dimension of the plate.

    Uses constants from NACA TN3781.

    Args:
        length: The length of the plate, in m.

        width: The width of the plate, in m.

        wall_thickness: The wall thickness of the plate, in m.

        elastic_modulus: The elastic modulus of the material, in Pa.

        side_boundary_condition_type: The boundary condition type at the sides of the plate. Options are:
            - "clamp-clamp"
            - "pin-pin"
            - "free-free"

    Returns:
        The critical compressive load (in N) for the plate to buckle via plate buckling.

    """

    if side_boundary_condition_type == "clamp-clamp":
        K = 6.35  # From NACA TN3781
    elif side_boundary_condition_type == "pin-pin":
        K = 3.62  # From NACA TN3781
    elif side_boundary_condition_type == "free-free":
        K = 0.385  # From NACA TN3781
    else:
        raise ValueError("Invalid `side_boundary_condition_type`.")

    critical_buckling_stress = K * elastic_modulus * (wall_thickness / width) ** 2

    plate_xsec_area = wall_thickness * width

    critical_buckling_load = critical_buckling_stress * plate_xsec_area

    return critical_buckling_load



import aerosandbox as asb
import aerosandbox.numpy as np
from typing import List
import copy
import pytest


def test_alpha_wind():
    op_point = asb.OperatingPoint(
        alpha=90,
        beta=0
    )
    x, y, z = op_point.convert_axes(
        0, 0, 1,
        "geometry",
        "wind"
    )
    assert x == pytest.approx(-1)
    assert y == pytest.approx(0)
    assert z == pytest.approx(0)


def test_beta_wind():
    op_point = asb.OperatingPoint(
        alpha=0,
        beta=90
    )
    x, y, z = op_point.convert_axes(
        0, 1, 0,
        "geometry",
        "wind"
    )
    assert x == pytest.approx(1)
    assert y == pytest.approx(0)
    assert z == pytest.approx(0)


def test_beta_wind_body():
    op_point = asb.OperatingPoint(
        alpha=0,
        beta=90
    )
    x, y, z = op_point.convert_axes(
        0, 1, 0,
        "body",
        "wind"
    )
    assert x == pytest.approx(1)
    assert y == pytest.approx(0)
    assert z == pytest.approx(0)


def test_alpha_stability_body():
    op_point = asb.OperatingPoint(
        alpha=90,
        beta=0
    )
    x, y, z = op_point.convert_axes(
        0, 0, 1,
        "body",
        "stability"
    )
    assert x == pytest.approx(1)
    assert y == pytest.approx(0)
    assert z == pytest.approx(0)


def test_beta_stability_body():
    op_point = asb.OperatingPoint(
        alpha=0,
        beta=90
    )
    x, y, z = op_point.convert_axes(
        0, 1, 0,
        "body",
        "stability"
    )
    assert x == pytest.approx(0)
    assert y == pytest.approx(1)
    assert z == pytest.approx(0)


def test_order_wind_body():
    op_point = asb.OperatingPoint(
        alpha=90,
        beta=90,
    )
    x, y, z = op_point.convert_axes(
        0, 1, 0,
        "body",
        "wind"
    )
    assert x == pytest.approx(1)
    assert y == pytest.approx(0)
    assert z == pytest.approx(0)


if __name__ == '__main__':
    pytest.main()



import aerosandbox.numpy as np
from typing import Union


def calculate_induced_velocity_point_source(
        x_field: Union[float, np.ndarray],
        y_field: Union[float, np.ndarray],
        z_field: Union[float, np.ndarray],
        x_source: Union[float, np.ndarray],
        y_source: Union[float, np.ndarray],
        z_source: Union[float, np.ndarray],
        sigma: Union[float, np.ndarray] = 1,
        viscous_radius=0,
) -> [Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculates the induced velocity at a point:
        [x_field, y_field, z_field]
    in a 3D potential-flow flowfield.

    In this flowfield, the following singularity elements are assumed:
        * A single point source

    This function consists entirely of scalar, elementwise NumPy ufunc operations - so it can be vectorized as
    desired assuming input dimensions/broadcasting are compatible.

    Args:
        x_field: x-coordinate of the field point

        y_field: y-coordinate of the field point

        z_field: z-coordinate of the field point

        x_left: x-coordinate of the left vertex of the bound vortex

        y_left: y-coordinate of the left vertex of the bound vortex

        z_left: z-coordinate of the left vertex of the bound vortex

        x_right: x-coordinate of the right vertex of the bound vortex

        y_right: y-coordinate of the right vertex of the bound vortex

        z_right: z-coordinate of the right vertex of the bound vortex

        gamma: The strength of the horseshoe vortex filament.

        trailing_vortex_direction: The direction that the trailing legs of the horseshoe vortex extend. Usually,
        this is modeled as the direction of the freestream.

        viscous_radius: To prevent a vortex singularity, here we use a Kaufmann vortex model. This parameter
        governs the radius of this vortex model. It should be significantly smaller (e.g., at least an order of
        magnitude smaller) than the smallest bound leg in the analysis in question.

    Returns: u, v, and w:
        The x-, y-, and z-direction induced velocities.
    """
    dx = x_field - x_source
    dy = y_field - y_source
    dz = z_field - z_source

    r_squared = (
            dx ** 2 +
            dy ** 2 +
            dz ** 2
    )

    def smoothed_x_15_inv(x):
        """
        Approximates x^(-1.5) with a function that sharply goes to 0 in the x -> 0 limit.
        """
        if not np.all(viscous_radius == 0):
            return x / (x ** 2.5 + viscous_radius ** 2.5)
        else:
            return x ** -1.5

    grad_phi_multiplier = sigma * smoothed_x_15_inv(r_squared) / (4 * np.pi)

    u = grad_phi_multiplier * dx
    v = grad_phi_multiplier * dy
    w = grad_phi_multiplier * dz

    return u, v, w


if __name__ == '__main__':
    args = (-2, 2, 30)
    x = np.linspace(*args)
    y = np.linspace(*args)
    z = np.linspace(*args)
    X, Y, Z = np.meshgrid(x, y, z)

    Xf = X.flatten()
    Yf = Y.flatten()
    Zf = Z.flatten()


    def wide(array):
        return np.reshape(array, (1, -1))


    def tall(array):
        return np.reshape(array, (-1, 1))


    Uf, Vf, Wf = calculate_induced_velocity_point_source(
        x_field=Xf,
        y_field=Yf,
        z_field=Zf,
        x_source=1,
        y_source=0,
        z_source=0,
    )

    pos = np.stack((Xf, Yf, Zf)).T
    dir = np.stack((Uf, Vf, Wf)).T

    dir_norm = np.reshape(np.linalg.norm(dir, axis=1), (-1, 1))

    dir = dir / dir_norm * dir_norm ** 0.2

    import pyvista as pv

    pv.set_plot_theme('dark')
    plotter = pv.Plotter()
    plotter.add_arrows(
        cent=pos,
        direction=dir,
        mag=0.15
    )
    plotter.show_grid()
    plotter.show()



import aerosandbox.numpy as np
from aerosandbox.common import AeroSandboxObject
from typing import Union, Any, List
from aerosandbox.tools.string_formatting import trim_string


class MassProperties(AeroSandboxObject):
    """
    Mass properties of a rigid 3D object.

    ## Notes on Inertia Tensor Definition

    This class uses the standard mathematical definition of the inertia tensor, which is different from the
    alternative definition used by some CAD and CAE applications (such as SolidWorks, NX, etc.). These differ by a
    sign flip in the products of inertia.

    Specifically, we define the inertia tensor using the standard convention:

        [ I11  I12  I13 ]   [ Ixx  Ixy  Ixz ]   [sum(m*(y^2+z^2))  -sum(m*x*y)      -sum(m*x*z)      ]
    I = [ I21  I22  I23 ] = [ Ixy  Iyy  Iyz ] = [-sum(m*x*y)       sum(m*(x^2+z^2)) -sum(m*y*z)      ]
        [ I31  I32  I33 ]   [ Ixz  Iyz  Izz ]   [-sum(m*x*z)       -sum(m*y*z)       sum(m*(x^2+y^2))]

    Whereas SolidWorks, NX, etc. define the inertia tensor as:

        [ I11  I12  I13 ]   [ Ixx -Ixy -Ixz ]   [sum(m*(y^2+z^2))  -sum(m*x*y)      -sum(m*x*z)      ]
    I = [ I21  I22  I23 ] = [-Ixy  Iyy -Iyz ] = [-sum(m*x*y)       sum(m*(x^2+z^2)) -sum(m*y*z)      ]
        [ I31  I32  I33 ]   [-Ixz -Iyz  Izz ]   [-sum(m*x*z)       -sum(m*y*z)       sum(m*(x^2+y^2))]

    See also: https://en.wikipedia.org/wiki/Moment_of_inertia#Inertia_tensor

    """

    def __init__(self,
                 mass: Union[float, np.ndarray] = None,
                 x_cg: Union[float, np.ndarray] = 0.,
                 y_cg: Union[float, np.ndarray] = 0.,
                 z_cg: Union[float, np.ndarray] = 0.,
                 Ixx: Union[float, np.ndarray] = 0.,
                 Iyy: Union[float, np.ndarray] = 0.,
                 Izz: Union[float, np.ndarray] = 0.,
                 Ixy: Union[float, np.ndarray] = 0.,
                 Iyz: Union[float, np.ndarray] = 0.,
                 Ixz: Union[float, np.ndarray] = 0.,
                 ):
        """
        Initializes a new MassProperties object.

        Axes can be given in any convenient axes system, as long as mass properties are not combined across different
        axis systems. For aircraft design, the most common axis system is typically geometry axes (x-positive aft,
        y-positive out the right wingtip, z-positive upwards).

        Args:

            mass: Mass of the component [kg]

            x_cg: X-location of the center of gravity of the component [m]

            y_cg: Y-location of the center of gravity of the component [m]

            z_cg:  Z-location of the center of gravity of the component [m]

            Ixx: Respective component of the inertia tensor, as measured about the component's center of mass. 0 if
            this is a point mass.

            Iyy: Respective component of the inertia tensor, as measured about the component's center of mass. 0 if
            this is a point mass.

            Izz: Respective component of the inertia tensor, as measured about the component's center of mass. 0 if
            this is a point mass.

            Ixy: Respective component of the inertia tensor, as measured about the component's center of mass. 0 if
            this is symmetric about z.

            Iyz: Respective component of the inertia tensor, as measured about the component's center of mass. 0 if
            this is symmetric about x.

            Ixz: Respective component of the inertia tensor, as measured about the component's center of mass. 0 if
            this is symmetric about y.


        """
        if mass is None:
            import warnings
            warnings.warn(
                "Defining a MassProperties object with zero mass. This can cause problems (divide-by-zero) in dynamics calculations, if this is not intended.\nTo silence this warning, please explicitly set `mass=0` in the MassProperties constructor.",
                stacklevel=2
            )
            mass = 0

        self.mass = mass
        self.x_cg = x_cg
        self.y_cg = y_cg
        self.z_cg = z_cg
        self.Ixx = Ixx
        self.Iyy = Iyy
        self.Izz = Izz
        self.Ixy = Ixy
        self.Iyz = Iyz
        self.Ixz = Ixz

    def __repr__(self) -> str:

        def fmt(x: Union[float, Any], width=14) -> str:
            if isinstance(x, (float, int)):
                if x == 0:
                    x = "0"
                else:
                    return f"{x:.8g}".rjust(width)
            return trim_string(str(x).rjust(width), length=40)

        return "\n".join([
            "MassProperties instance:",
            f"                 Mass : {fmt(self.mass)}",
            f"    Center of Gravity : ({fmt(self.x_cg)}, {fmt(self.y_cg)}, {fmt(self.z_cg)})",
            f"       Inertia Tensor : ",
            f"            (about CG)  [{fmt(self.Ixx)}, {fmt(self.Ixy)}, {fmt(self.Ixz)}]",
            f"                        [{fmt(self.Ixy)}, {fmt(self.Iyy)}, {fmt(self.Iyz)}]",
            f"                        [{fmt(self.Ixz)}, {fmt(self.Iyz)}, {fmt(self.Izz)}]",
        ])

    def __getitem__(self, index):
        def get_item_of_attribute(a):
            if np.isscalar(a):  # If NumPy says its a scalar, return it.
                return a

            try:
                return a[index]
            except TypeError:  # object is not subscriptable
                return a
            except IndexError as e:  # index out of range
                raise IndexError("A state variable could not be indexed, since the index is out of range!")

        inputs = {
            "mass": self.mass,
            "x_cg": self.x_cg,
            "y_cg": self.y_cg,
            "z_cg": self.z_cg,
            "Ixx" : self.Ixx,
            "Iyy" : self.Iyy,
            "Izz" : self.Izz,
            "Ixy" : self.Ixy,
            "Iyz" : self.Iyz,
            "Ixz" : self.Ixz,
        }

        return self.__class__(
            **{
                k: get_item_of_attribute(v)
                for k, v in inputs.items()
            }
        )

    def __neg__(self):
        return -1 * self

    def __add__(self, other: "MassProperties") -> "MassProperties":
        """
        Combines one MassProperties object with another.
        """
        total_mass = self.mass + other.mass
        total_x_cg = (self.mass * self.x_cg + other.mass * other.x_cg) / total_mass
        total_y_cg = (self.mass * self.y_cg + other.mass * other.y_cg) / total_mass
        total_z_cg = (self.mass * self.z_cg + other.mass * other.z_cg) / total_mass
        self_inertia_tensor_elements = self.get_inertia_tensor_about_point(
            x=total_x_cg,
            y=total_y_cg,
            z=total_z_cg,
            return_tensor=False
        )
        other_inertia_tensor_elements = other.get_inertia_tensor_about_point(
            x=total_x_cg,
            y=total_y_cg,
            z=total_z_cg,
            return_tensor=False
        )

        total_inertia_tensor_elements = [
            I__ + J__
            for I__, J__ in zip(
                self_inertia_tensor_elements,
                other_inertia_tensor_elements
            )
        ]

        return MassProperties(
            mass=total_mass,
            x_cg=total_x_cg,
            y_cg=total_y_cg,
            z_cg=total_z_cg,
            Ixx=total_inertia_tensor_elements[0],
            Iyy=total_inertia_tensor_elements[1],
            Izz=total_inertia_tensor_elements[2],
            Ixy=total_inertia_tensor_elements[3],
            Iyz=total_inertia_tensor_elements[4],
            Ixz=total_inertia_tensor_elements[5],
        )

    def __radd__(self, other):
        """
        Allows sum() to work with MassProperties objects.

        Basically, makes addition commutative.
        """
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __sub__(self, other: "MassProperties") -> "MassProperties":
        """
        Subtracts one MassProperties object from another. (opposite of __add__() )
        """
        new_mass = self.mass - other.mass
        new_x_cg = (self.x_cg * self.mass - other.x_cg * other.mass) / (self.mass - other.mass)
        new_y_cg = (self.y_cg * self.mass - other.y_cg * other.mass) / (self.mass - other.mass)
        new_z_cg = (self.z_cg * self.mass - other.z_cg * other.mass) / (self.mass - other.mass)
        self_inertia_tensor_elements = self.get_inertia_tensor_about_point(
            x=new_x_cg,
            y=new_y_cg,
            z=new_z_cg,
            return_tensor=False
        )
        other_inertia_tensor_elements = other.get_inertia_tensor_about_point(
            x=new_x_cg,
            y=new_y_cg,
            z=new_z_cg,
            return_tensor=False
        )

        new_inertia_tensor_elements = [
            I__ - J__
            for I__, J__ in zip(
                self_inertia_tensor_elements,
                other_inertia_tensor_elements
            )
        ]

        return MassProperties(
            mass=new_mass,
            x_cg=new_x_cg,
            y_cg=new_y_cg,
            z_cg=new_z_cg,
            Ixx=new_inertia_tensor_elements[0],
            Iyy=new_inertia_tensor_elements[1],
            Izz=new_inertia_tensor_elements[2],
            Ixy=new_inertia_tensor_elements[3],
            Iyz=new_inertia_tensor_elements[4],
            Ixz=new_inertia_tensor_elements[5],
        )

    def __mul__(self, other: float) -> "MassProperties":
        """
        Returns a new MassProperties object that is equivalent to if you had summed together N (with `other`
        interpreted as N) identical MassProperties objects.
        """
        return MassProperties(
            mass=self.mass * other,
            x_cg=self.x_cg,
            y_cg=self.y_cg,
            z_cg=self.z_cg,
            Ixx=self.Ixx * other,
            Iyy=self.Iyy * other,
            Izz=self.Izz * other,
            Ixy=self.Ixy * other,
            Iyz=self.Iyz * other,
            Ixz=self.Ixz * other,
        )

    def __rmul__(self, other: float) -> "MassProperties":
        """
        Allows multiplication of a scalar by a MassProperties object.
        """
        return self.__mul__(other)

    def __truediv__(self, other: float) -> "MassProperties":
        """
        Returns a new MassProperties object that is equivalent to if you had divided the mass of the current
        MassProperties object by a factor.
        """
        return self.__mul__(1 / other)

    def __eq__(self, other: "MassProperties") -> bool:
        """
        Returns True if all expected attributes of the two MassProperties objects are exactly equal.
        """
        return all([
            getattr(self, attribute) == getattr(other, attribute)
            for attribute in [
                "mass",
                "x_cg",
                "y_cg",
                "z_cg",
                "Ixx",
                "Iyy",
                "Izz",
                "Ixy",
                "Iyz",
                "Ixz",
            ]
        ])

    def __ne__(self, other: "MassProperties") -> bool:
        return not self.__eq__(other)

    def allclose(self,
                 other: "MassProperties",
                 rtol=1e-5,
                 atol=1e-8,
                 equal_nan=False
                 ) -> bool:
        return all([
            np.allclose(
                getattr(self, attribute),
                getattr(other, attribute),
                rtol=rtol,
                atol=atol,
                equal_nan=equal_nan
            )
            for attribute in [
                "mass",
                "x_cg",
                "y_cg",
                "z_cg",
                "Ixx",
                "Iyy",
                "Izz",
                "Ixy",
                "Iyz",
                "Ixz",
            ]
        ])

    @property
    def xyz_cg(self):
        return self.x_cg, self.y_cg, self.z_cg

    @property
    def inertia_tensor(self):
        # Returns the inertia tensor about the component's centroid.
        return np.array(
            [[self.Ixx, self.Ixy, self.Ixz],
             [self.Ixy, self.Iyy, self.Iyz],
             [self.Ixz, self.Iyz, self.Izz]]
        )

    def inv_inertia_tensor(self):
        """
        Computes the inverse of the inertia tensor, in a slightly more efficient way than raw inversion by exploiting its known structure.

        If you are effectively using this inertia tensor to solve a linear system, you should use a linear algebra
        solve() method (ideally via Cholesky decomposition) instead, for best speed.
        """
        iIxx, iIyy, iIzz, iIxy, iIyz, iIxz = np.linalg.inv_symmetric_3x3(
            m11=self.Ixx,
            m22=self.Iyy,
            m33=self.Izz,
            m12=self.Ixy,
            m23=self.Iyz,
            m13=self.Ixz,
        )
        return np.array(
            [[iIxx, iIxy, iIxz],
             [iIxy, iIyy, iIyz],
             [iIxz, iIyz, iIzz]]
        )

    def get_inertia_tensor_about_point(self,
                                       x: float = 0.,
                                       y: float = 0.,
                                       z: float = 0.,
                                       return_tensor: bool = True,
                                       ):
        """
        Returns the inertia tensor about an arbitrary point.
        Using https://en.wikipedia.org/wiki/Parallel_axis_theorem#Tensor_generalization

        Args:
            x: x-position of the new point, in the same axes as this MassProperties instance is specified in.

            y: y-position of the new point, in the same axes as this MassProperties instance is specified in.

            z: z-position of the new point, in the same axes as this MassProperties instance is specified in.

            return_tensor: A switch for the desired return type; see below for details. [boolean]

        Returns:

            If `return_tensor` is True:
                Returns the new inertia tensor, as a 2D numpy ndarray.
            If `return_tensor` is False:
                Returns the components of the new inertia tensor, as a tuple.
                If J is the new inertia tensor, the tuple returned is:
                (Jxx, Jyy, Jzz, Jxy, Jyz, Jxz)

        """

        R = [x - self.x_cg, y - self.y_cg, z - self.z_cg]
        RdotR = np.dot(R, R, manual=True)

        Jxx = self.Ixx + self.mass * (RdotR - R[0] ** 2)
        Jyy = self.Iyy + self.mass * (RdotR - R[1] ** 2)
        Jzz = self.Izz + self.mass * (RdotR - R[2] ** 2)
        Jxy = self.Ixy - self.mass * R[0] * R[1]
        Jyz = self.Iyz - self.mass * R[1] * R[2]
        Jxz = self.Ixz - self.mass * R[2] * R[0]

        if return_tensor:
            return np.array([
                [Jxx, Jxy, Jxz],
                [Jxy, Jyy, Jyz],
                [Jxz, Jyz, Jzz],
            ])
        else:
            return Jxx, Jyy, Jzz, Jxy, Jyz, Jxz

    def is_physically_possible(self) -> bool:
        """
        Checks whether it's possible for this MassProperties object to correspond to the mass properties of a real
        physical object.

        Assumes that all physically-possible objects have a positive mass (or density).

        Some special edge cases:

            - A MassProperties object with mass of 0 (i.e., null object) will return True. Note: this will return
            True even if the inertia tensor is not zero (which would basically be infinitesimal point masses at
            infinite distance).

            - A MassProperties object that is a point mass (i.e., inertia tensor is all zeros) will return True.

        Returns:
            True if the MassProperties object is physically possible, False otherwise.
        """

        ### This checks the basics
        impossible_conditions = [
            self.mass < 0,
            self.Ixx < 0,
            self.Iyy < 0,
            self.Izz < 0,
        ]

        eigs = np.linalg.eig(self.inertia_tensor)[0]

        # ## This checks that the inertia tensor is positive definite, which is a necessary but not sufficient
        # condition for an inertia tensor to be physically possible.
        impossible_conditions.extend([
            eigs[0] < 0,
            eigs[1] < 0,
            eigs[2] < 0,
        ])

        # ## This checks the triangle inequality, which is a necessary but not sufficient condition for an inertia
        # tensor to be physically possible.
        impossible_conditions.extend([
            eigs[0] + eigs[1] < eigs[2],
            eigs[0] + eigs[2] < eigs[1],
            eigs[1] + eigs[2] < eigs[0],
        ])

        return not any(impossible_conditions)

    def is_point_mass(self) -> bool:
        """
        Returns True if this MassProperties object corresponds to a point mass, False otherwise.
        """
        return np.allclose(self.inertia_tensor, 0)

    def generate_possible_set_of_point_masses(self,
                                              method="optimization",
                                              check_if_already_a_point_mass: bool = True,
                                              ) -> List["MassProperties"]:
        """
        Generates a set of point masses (represented as MassProperties objects with zero inertia tensors), that, when
        combined, would yield this MassProperties object.

        Note that there are an infinite number of possible sets of point masses that could yield this MassProperties
        object. This method returns one possible set of point masses, but there are many others.

        Example:
            >>> mp = MassProperties(mass=1, Ixx=1, Iyy=1, Izz=1, Ixy=0.1, Iyz=-0.1, Ixz=0.1)
            >>> point_masses = mp.generate_possible_set_of_point_masses()
            >>> mp.allclose(sum(point_masses))  # Asserts these are equal, within tolerance
            True

        Args:
            method: The method to use to generate the set of point masses. Currently, only "barbell" is supported.

        Returns:
            A list of MassProperties objects, each of which is a point mass (i.e., zero inertia tensor).
        """
        if check_if_already_a_point_mass:
            if self.is_point_mass():
                return [self]

        if method == "optimization":
            from aerosandbox.optimization import Opti

            opti = Opti()

            approximate_radius = (self.Ixx + self.Iyy + self.Izz) ** 0.5 / self.mass + 1e-16

            point_masses = [
                MassProperties(
                    mass=self.mass / 4,
                    x_cg=opti.variable(init_guess=self.x_cg - approximate_radius, scale=approximate_radius),
                    y_cg=opti.variable(init_guess=self.y_cg, scale=approximate_radius),
                    z_cg=opti.variable(init_guess=self.z_cg, scale=approximate_radius),
                ),
                MassProperties(
                    mass=self.mass / 4,
                    x_cg=opti.variable(init_guess=self.x_cg, scale=approximate_radius),
                    y_cg=opti.variable(init_guess=self.y_cg, scale=approximate_radius),
                    z_cg=opti.variable(init_guess=self.z_cg + approximate_radius, scale=approximate_radius),
                ),
                MassProperties(
                    mass=self.mass / 4,
                    x_cg=opti.variable(init_guess=self.x_cg, scale=approximate_radius),
                    y_cg=opti.variable(init_guess=self.y_cg, scale=approximate_radius),
                    z_cg=opti.variable(init_guess=self.z_cg - approximate_radius, scale=approximate_radius),
                ),
                MassProperties(
                    mass=self.mass / 4,
                    x_cg=opti.variable(init_guess=self.x_cg, scale=approximate_radius),
                    y_cg=opti.variable(init_guess=self.y_cg + approximate_radius, scale=approximate_radius),
                    z_cg=opti.variable(init_guess=self.z_cg, scale=approximate_radius),
                ),
            ]

            mass_props_reconstructed = sum(point_masses)

            # Add constraints
            opti.subject_to(mass_props_reconstructed.x_cg == self.x_cg)
            opti.subject_to(mass_props_reconstructed.y_cg == self.y_cg)
            opti.subject_to(mass_props_reconstructed.z_cg == self.z_cg)
            opti.subject_to(mass_props_reconstructed.Ixx == self.Ixx)
            opti.subject_to(mass_props_reconstructed.Iyy == self.Iyy)
            opti.subject_to(mass_props_reconstructed.Izz == self.Izz)
            opti.subject_to(mass_props_reconstructed.Ixy == self.Ixy)
            opti.subject_to(mass_props_reconstructed.Iyz == self.Iyz)
            opti.subject_to(mass_props_reconstructed.Ixz == self.Ixz)

            opti.subject_to(point_masses[0].y_cg == self.y_cg)
            opti.subject_to(point_masses[0].z_cg == self.z_cg)
            opti.subject_to(point_masses[1].y_cg == self.y_cg)

            opti.subject_to(point_masses[0].x_cg < point_masses[1].x_cg)

            return opti.solve(verbose=False)(point_masses)


        elif method == "barbell":
            raise NotImplementedError("Barbell method not yet implemented!")
            principle_inertias, principle_axes = np.linalg.eig(self.inertia_tensor)

        else:
            raise ValueError("Bad value of `method` argument!")

    def export_AVL_mass_file(self,
                             filename,
                             ) -> None:
        """
        Exports this MassProperties object to an AVL mass file.

        Note: AVL uses the SolidWorks convention for inertia tensors, which is different from the typical
        mathematical convention, and the convention used by this MassProperties class. In short, these differ by a
        sign flip in the products of inertia. More details available in the MassProperties docstring. See also:
        https://en.wikipedia.org/wiki/Moment_of_inertia#Inertia_tensor

        Args:
            filename: The filename to export to.

        Returns: None

        """
        lines = [
            "Lunit = 1.0 m",
            "Munit = 1.0 kg",
            "Tunit = 1.0 s",
            "",
            "g     = 9.81",
            "rho   = 1.225",
            "",
        ]

        def fmt(x: float) -> str:
            return f"{x:.16g}".ljust(21)

        lines.extend([
            " ".join([
                s.ljust(21) for s in [
                    "#  mass",
                    "x_cg",
                    "y_cg",
                    "z_cg",
                    "Ixx",
                    "Iyy",
                    "Izz",
                    "Ixy",
                    "Ixz",
                    "Iyz",
                ]
            ]),
            " ".join([
                fmt(x) for x in [
                    self.mass,
                    self.x_cg,
                    self.y_cg,
                    self.z_cg,
                    self.Ixx,
                    self.Iyy,
                    self.Izz,
                    -self.Ixy,
                    -self.Ixz,
                    -self.Iyz,
                ]
            ])
        ])

        with open(filename, "w+") as f:
            f.write("\n".join(lines))


if __name__ == '__main__':
    mp1 = MassProperties(
        mass=1
    )
    mp2 = MassProperties(
        mass=1,
        x_cg=1
    )
    mps = mp1 + mp2
    assert mps.x_cg == 0.5

    assert mp1 + mp2 - mp2 == mp1

    r = lambda: np.random.randn()

    valid = False
    while not valid:
        mass_props = MassProperties(
            mass=r(),
            x_cg=r(), y_cg=r(), z_cg=r(),
            Ixx=r(), Iyy=r(), Izz=r(),
            Ixy=r(), Iyz=r(), Ixz=r(),
        )
        valid = mass_props.is_physically_possible()  # adds a bunch of checks



import aerosandbox.numpy as np
from aerosandbox.tools import units as u
from typing import Union, Dict


def motor_electric_performance(
        voltage: Union[float, np.ndarray] = None,
        current: Union[float, np.ndarray] = None,
        rpm: Union[float, np.ndarray] = None,
        torque: Union[float, np.ndarray] = None,
        kv: float = 1000.,  # rpm/volt
        resistance: float = 0.1,  # ohms
        no_load_current: float = 0.4  # amps
) -> Dict[str, Union[float, np.ndarray]]:
    """
    A function for predicting the performance of an electric motor.
    
    Performance equations based on Mark Drela's First Order Motor Model:
    http://web.mit.edu/drela/Public/web/qprop/motor1_theory.pdf
    
    Instructions: Input EXACTLY TWO of the following parameters: voltage, current, rpm, torque.
    
    Exception: You cannot supply the combination of current and torque - this makes for an ill-posed problem.
    
    Note that this function is fully vectorized, so arrays can be supplied to any of the inputs.
    
    Args:
        voltage: Voltage across motor terminals [Volts]
        
        current: Current through motor [Amps]
        
        rpm: Motor rotation speed [rpm]
        
        torque: Motor torque [N m]
        
        kv: voltage constant, in rpm/volt
        
        resistance: resistance, in ohms
        
        no_load_current: no-load current, in amps
        
    Returns:
        A dictionary where keys are: 
            "voltage", 
            "current", 
            "rpm", 
            "torque",
            "shaft power", 
            "electrical power", 
            "efficiency"
            "waste heat"

        And values are corresponding quantities in SI units.

        Note that "efficiency" is just (shaft power) / (electrical power), and hence implicitly assumes that the
        motor is operating as a motor (electrical -> shaft power), and not a generator (shaft power -> electrical).
        If you want to know the efficiency of the motor as a generator, you can simply calculate it as (electrical
        power) / (shaft power).
    """
    # Validate inputs
    voltage_known = voltage is not None
    current_known = current is not None
    rpm_known = rpm is not None
    torque_known = torque is not None

    if not (
            voltage_known + current_known + rpm_known + torque_known == 2
    ):
        raise ValueError("You must give exactly two input arguments.")

    if current_known and torque_known:
        raise ValueError(
            "You cannot supply the combination of current and torque - this makes for an ill-posed problem.")

    kv_rads_per_sec_per_volt = kv * np.pi / 30  # rads/sec/volt

    ### Iterate through the motor equations until all quantities are known.
    while not (voltage_known and current_known and rpm_known and torque_known):
        if rpm_known:
            if current_known and not voltage_known:
                speed = rpm * np.pi / 30  # rad/sec
                back_EMF_voltage = speed / kv_rads_per_sec_per_volt
                voltage = back_EMF_voltage + current * resistance
                voltage_known = True

        if torque_known:
            if not current_known:
                current = torque * kv_rads_per_sec_per_volt + no_load_current
                current_known = True

        if voltage_known:
            if rpm_known and not current_known:
                speed = rpm * np.pi / 30  # rad/sec
                back_EMF_voltage = speed / kv_rads_per_sec_per_volt
                current = (voltage - back_EMF_voltage) / resistance
                current_known = True
            if not rpm_known and current_known:
                back_EMF_voltage = voltage - (current * resistance)
                speed = back_EMF_voltage * kv_rads_per_sec_per_volt
                rpm = speed * 30 / np.pi
                rpm_known = True

        if current_known:
            if not torque_known:
                torque = (current - no_load_current) / kv_rads_per_sec_per_volt
                torque_known = True

    shaft_power = (rpm * np.pi / 30) * torque
    electrical_power = voltage * current
    efficiency = shaft_power / electrical_power
    waste_heat = np.fabs(electrical_power - shaft_power)

    return {
        "voltage"         : voltage,
        "current"         : current,
        "rpm"             : rpm,
        "torque"          : torque,
        "shaft power"     : shaft_power,
        "electrical power": electrical_power,
        "efficiency"      : efficiency,
        "waste heat"      : waste_heat,
    }


def motor_resistance_from_no_load_current(
        no_load_current
):
    """
    Estimates the internal resistance of a motor from its no_load_current. Gates quotes R^2=0.93 for this model.

    Source: Gates, et. al., "Combined Trajectory, Propulsion, and Battery Mass Optimization for Solar-Regen..."
        https://scholarsarchive.byu.edu/cgi/viewcontent.cgi?article=3932&context=facpub

    Args:
        no_load_current: No-load current [amps]

    Returns:
        motor internal resistance [ohms]
    """
    return 0.0467 * no_load_current ** -1.892


def mass_ESC(
        max_power,
):
    """
    Estimates the mass of an ESC.

    Informal correlation I did to Hobbyking ESCs in the 8S LiPo, 100A range

    Args:
        max_power: maximum power [W]

    Returns:
        estimated ESC mass [kg]
    """
    return 2.38e-5 * max_power


def mass_battery_pack(
        battery_capacity_Wh,
        battery_cell_specific_energy_Wh_kg=240,
        battery_pack_cell_fraction=0.7,
):
    """
    Estimates the mass of a lithium-polymer battery.

    Args:
        battery_capacity_Wh: Battery capacity, in Watt-hours [W*h]

        battery_cell_specific_energy: Specific energy of the battery at the CELL level [W*h/kg]

        battery_pack_cell_fraction: Fraction of the battery pack that is cells, by weight.

            * Note: Ed Lovelace, a battery engineer for Aurora Flight Sciences, gives this figure as 0.70 in a Feb.
            2020 presentation for MIT 16.82

    Returns:
        Estimated battery mass [kg]
    """
    return battery_capacity_Wh / battery_cell_specific_energy_Wh_kg / battery_pack_cell_fraction


def mass_motor_electric(
        max_power,
        kv_rpm_volt=1000,  # This is in rpm/volt, not rads/sec/volt!
        voltage=20,
        method="hobbyking"
):
    """
    Estimates the mass of a brushless DC electric motor.
    Curve fit to scraped Hobbyking BLDC motor data as of 2/24/2020.
    Estimated range of validity: 50 < max_power < 10000

    Args:
        max_power (float): maximum power [W]

        kv_rpm_volt (float): Voltage constant of the motor, measured in rpm/volt, not rads/sec/volt! [rpm/volt]

        voltage (float): Operating voltage of the motor [V]

        method (str): method to use. "burton", "hobbyking", or "astroflight" (increasing level of detail).

            * Burton source: https://dspace.mit.edu/handle/1721.1/112414

            * Hobbyking source: C:\Projects\GitHub\MotorScraper,

            * Astroflight source: Gates, et. al., "Combined Trajectory, Propulsion, and Battery Mass Optimization for
            Solar-Regen..." https://scholarsarchive.byu.edu/cgi/viewcontent.cgi?article=3932&context=facpub

                * Validity claimed from 1.5 kW to 15 kW, kv from 32 to 1355.

    Returns:
        Estimated motor mass [kg]
    """
    if method == "burton":
        return max_power / 4128  # Less sophisticated model. 95% CI (3992, 4263), R^2 = 0.866
    elif method == "hobbyking":
        return 10 ** (0.8205 * np.log10(max_power) - 3.155)  # More sophisticated model
    elif method == "astroflight":
        max_current = max_power / voltage
        return 2.464 * max_current / kv_rpm_volt + 0.368  # Even more sophisticated model


def mass_wires(
        wire_length,
        max_current,
        allowable_voltage_drop,
        material="aluminum",
        insulated=True,
        max_voltage=600,
        wire_packing_factor=1,
        insulator_density=1700,
        insulator_dielectric_strength=12e6,
        insulator_min_thickness=0.2e-3,  # silicone wire
        return_dict: bool = False
):
    """
    Estimates the mass of wires used for power transmission.

    Materials data from: https://en.wikipedia.org/wiki/Electrical_resistivity_and_conductivity#Resistivity-density_product
        All data measured at STP; beware, as this data (especially resistivity) can be a strong function of temperature.

    Args:
        wire_length (float): Length of the wire [m]

        max_current (float): Max current of the wire [Amps]

        allowable_voltage_drop (float): How much is the voltage allowed to drop along the wire?

        material (str): Conductive material of the wire ("aluminum"). Determines density and resistivity. One of:

            * "sodium"

            * "lithium"

            * "calcium"

            * "potassium"

            * "beryllium"

            * "aluminum"

            * "magnesium"

            * "copper"

            * "silver"

            * "gold"

            * "iron"

        insulated (bool): Should we add the mass of the wire's insulator coating? Usually you'll want to leave this True.

        max_voltage (float): Maximum allowable voltage (used for sizing insulator). 600 is a common off-the-shelf rating.

        wire_packing_factor (float): What fraction of the enclosed cross section is conductor? This is 1 for solid wire,
            and less for stranded wire.

        insulator_density (float): Density of the insulator [kg/m^3]

        insulator_dielectric_strength (float): Dielectric strength of the insulator [V/m]. The default value of 12e6 corresponds
        to rubber.

        insulator_min_thickness (float): Minimum thickness of the insulator [m]. This is essentially a gauge limit.
        The default value is 0.2 mm.

        return_dict (bool): If True, returns a dictionary of all local variables. If False, just returns the wire
        mass as a float. Defaults to False.


    Returns: If `return_dict` is False (default), returns the wire mass as a single number. If `return_dict` is True,
    returns a dictionary of all local variables.
    """
    if material == "sodium":  # highly reactive with water & oxygen, low physical strength
        density = 970  # kg/m^3
        resistivity = 47.7e-9  # ohm-meters
    elif material == "lithium":  # highly reactive with water & oxygen, low physical strength
        density = 530  # kg/m^3
        resistivity = 92.8e-9  # ohm-meters
    elif material == "calcium":  # highly reactive with water & oxygen, low physical strength
        density = 1550  # kg/m^3
        resistivity = 33.6e-9  # ohm-meters
    elif material == "potassium":  # highly reactive with water & oxygen, low physical strength
        density = 890  # kg/m^3
        resistivity = 72.0e-9  # ohm-meters
    elif material == "beryllium":  # toxic, brittle
        density = 1850  # kg/m^3
        resistivity = 35.6e-9  # ohm-meters
    elif material == "aluminum":
        density = 2700  # kg/m^3
        resistivity = 26.50e-9  # ohm-meters
    elif material == "magnesium":  # worse specific conductivity than aluminum
        density = 1740  # kg/m^3
        resistivity = 43.90e-9  # ohm-meters
    elif material == "copper":  # worse specific conductivity than aluminum, moderately expensive
        density = 8960  # kg/m^3
        resistivity = 16.78e-9  # ohm-meters
    elif material == "silver":  # worse specific conductivity than aluminum, expensive
        density = 10490  # kg/m^3
        resistivity = 15.87e-9  # ohm-meters
    elif material == "gold":  # worse specific conductivity than aluminum, very expensive
        density = 19300  # kg/m^3
        resistivity = 22.14e-9  # ohm-meters
    elif material == "iron":  # worse specific conductivity than aluminum
        density = 7874  # kg/m^3
        resistivity = 96.1e-9  # ohm-meters
    else:
        raise ValueError("Bad value of 'material'!")

    # Conductor mass
    resistance = allowable_voltage_drop / max_current
    area_conductor = resistivity * wire_length / resistance
    volume_conductor = area_conductor * wire_length
    mass_conductor = volume_conductor * density

    # Insulator mass
    if insulated:
        insulator_thickness = np.softmax(
            4.0 * max_voltage / insulator_dielectric_strength,
            insulator_min_thickness,
            softness=0.005 * u.inch,
        )
        radius_conductor = (area_conductor / wire_packing_factor / np.pi) ** 0.5
        radius_insulator = radius_conductor + insulator_thickness
        area_insulator = np.pi * radius_insulator ** 2 - area_conductor
        volume_insulator = area_insulator * wire_length
        mass_insulator = insulator_density * volume_insulator
    else:
        mass_insulator = 0

    # Total them up
    mass_total = mass_conductor + mass_insulator

    if return_dict:
        return locals()
    else:
        return mass_total


if __name__ == '__main__':
    print(motor_electric_performance(
        rpm=100,
        current=3
    ))
    print(motor_electric_performance(
        rpm=4700,
        torque=0.02482817
    ))

    print(
        mass_battery_pack(100)
    )

    pows = np.logspace(2, 5, 300)
    mass_mot_burton = mass_motor_electric(pows, method="burton")
    mass_mot_hobbyking = mass_motor_electric(pows, method="hobbyking")
    mass_mot_astroflight = mass_motor_electric(pows, method="astroflight")

    import matplotlib.pyplot as plt
    import aerosandbox.tools.pretty_plots as p

    fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
    plt.loglog(pows, np.array(mass_mot_burton), "-", label="Burton Model")
    plt.plot(pows, np.array(mass_mot_hobbyking), "--", label="Hobbyking Model")
    plt.plot(pows, np.array(mass_mot_astroflight), "-.", label="Astroflight Model")
    p.show_plot(
        "Small Electric Motor Mass Models\n(500 kv, 100 V)",
        "Motor Power [W]",
        "Motor Mass [kg]"
    )

    print(mass_wires(
        wire_length=1,
        max_current=100,
        allowable_voltage_drop=1,
        material="aluminum"
    ))



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
import codecs

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


### Set up tools to get version
def read(rel_path):
    here = path.abspath(path.dirname(__file__))
    with codecs.open(path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


### Do the setup
setup(
    name='AeroSandbox',
    author='Peter Sharpe',
    version=get_version("aerosandbox/__init__.py"),
    description='A Python 3 package for playing around with aerodynamics ideas related to vortex lattice methods, coupled viscous/inviscid methods, automatic differentiation for gradient computation, aircraft design optimization, and the like. Work in progress!',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://peterdsharpe.github.io/AeroSandbox/',
    author_email='pds@mit.edu',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    keywords='aerodynamics airplane cfd mdo mdao aircraft design aerospace optimization automatic differentiation',
    packages=find_packages(exclude=['docs', 'media', 'examples', 'studies']),
    python_requires='>=3.7',
    install_requires=[
        'numpy >= 1.20.0',
        'scipy >= 1.7.0',
        'casadi >= 3.5.5',
        'pandas >= 1',
        'matplotlib >= 3.3.1',
        'seaborn >= 0.11',
        'tqdm >= 4',
        'sortedcontainers > 2'
    ],
    extras_require={
        "full": [
            'plotly >= 5',
            'pyvista >= 0.31',
            'ipyvtklink >= 0.2',
            'trimesh >= 3',
            'sympy >= 1',
            'cadquery >= 2; python_version >= "3.8"',
        ],
        "test": [
            'pytest',
            'nbval'
        ],
        "docs": [
            'sphinx',
            'furo',
            'sphinx-autoapi',
        ],
    },
    include_package_data=True,
    package_data={
        'Airfoil database': ['*.dat'],  # include all airfoil *.dat files
    },
    project_urls={  # Optional
        'Source'     : 'https://github.com/peterdsharpe/AeroSandbox',
        'Bug Reports': 'https://github.com/peterdsharpe/AeroSandbox/issues',
    },
)
